@@ -1,4 +1,6 @@
-#!/bin/bash -e
+#!/bin/bash
+
+set -ex -o xtrace
 
 # install the opensc
 sudo make install
@@ -15,28 +17,26 @@ pushd libcacard
 ./autogen.sh --prefix=/usr && make -j2 && sudo make install
 popd
 
+# prepare pcscd
+. .github/restart-pcscd.sh
+
 # virt_cacard
 if [ ! -d "virt_cacard" ]; then
 	git clone https://github.com/Jakuje/virt_cacard.git
 fi
 pushd virt_cacard
 ./autogen.sh && ./configure && make
-popd
-
-sudo /etc/init.d/pcscd restart
-
-pushd src/tests/p11test/
-./p11test -s 0 -p 12345678 -i -o virt_cacard.json &
-sleep 5
-popd
-
-# virt_cacard startup
-pushd virt_cacard
 ./setup-softhsm2.sh
 export SOFTHSM2_CONF=$PWD/softhsm2.conf
-./virt_cacard &
-wait $(ps aux | grep '[p]11test'| awk '{print $2}')
-kill -9 $(ps aux | grep '[v]irt_cacard'| awk '{print $2}')
+# register cleanup function on exit
+trap "pkill -9 virt_cacard" EXIT
+./virt_cacard 2>&1 | sed -e 's/^/virt_cacard: /;' &
+popd
+
+# run the tests
+pushd src/tests/p11test/
+sleep 5
+./p11test -s 0 -p 12345678 -o virt_cacard.json
 popd
 
 # cleanup -- this would break later uses of pcscd
