@@ -1,4 +1,6 @@
-#!/bin/bash -e
+#!/bin/bash
+
+set -ex -o xtrace
 
 # install the opensc
 sudo make install
@@ -40,6 +42,9 @@ echo | ./OsEID-tool INIT
 ./OsEID-tool EC-SIGN-TEST
 ./OsEID-tool EC-SIGN-PKCS11-TEST
 ./OsEID-tool EC-ECDH-TEST
+./OsEID-tool UNWRAP-WRAP-TEST
+./OsEID-tool DES-AES-UPLOAD-KEYS
+./OsEID-tool SYM-CRYPT-TEST
 popd
 
 # this does not work as we have random key IDs in here
