@@ -111,7 +111,7 @@ static long t1, t2, tot_read = 0, tot_dur = 0, dur;
 /* Data in the return value for the GET CARD DATA command:
  * All fields are one byte, except when noted otherwise.
  *
- * See 6.9 in
+ * See §6.9 in
  * https://github.com/Fedict/eid-mw/blob/master/doc/sdk/documentation/Public_Belpic_Applet_v1%207_Ref_Manual%20-%20A01.pdf
  * for the full documentation on the GET CARD DATA command.
  */
@@ -146,6 +146,8 @@ static long t1, t2, tot_read = 0, tot_dur = 0, dur;
 static size_t next_idx = (size_t)-1;
 
 static const struct sc_atr_table belpic_atrs[] = {
+	/* Applet V1.8 */
+	{ "3B:7F:96:00:00:80:31:80:65:B0:85:04:01:20:12:0F:FF:82:90:00", NULL, NULL, SC_CARD_TYPE_BELPIC_EID, 0, NULL },
 	/* Applet V1.1 */
 	{ "3B:98:13:40:0A:A5:03:01:01:01:AD:13:11", NULL, NULL, SC_CARD_TYPE_BELPIC_EID, 0, NULL },
 	/* Applet V1.0 with new EMV-compatible ATR */
@@ -193,7 +195,7 @@ static int get_carddata(sc_card_t *card, u8* carddata_loc, unsigned int carddata
 		return r;
 	}
 	if(apdu.resplen < carddataloc_len) {
-		sc_log(card->ctx, 
+		sc_log(card->ctx,
 			 "GetCardData: card returned %"SC_FORMAT_LEN_SIZE_T"u bytes rather than expected %d\n",
 			 apdu.resplen, carddataloc_len);
 		return SC_ERROR_WRONG_LENGTH;
