@@ -1872,6 +1872,7 @@ static int cac_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data, int *tries
 	sc_apdu_t apdu;
 	u8  sbuf[SC_MAX_APDU_BUFFER_SIZE];
 	struct sc_card_driver *iso_drv = sc_get_iso7816_driver();
+	int rv;
 
 	if (data->cmd == SC_PIN_CMD_CHANGE) {
 		int i = 0;
@@ -1897,7 +1898,10 @@ static int cac_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data, int *tries
 		}
 	}
 
-	return  iso_drv->ops->pin_cmd(card, data, tries_left);
+	rv = iso_drv->ops->pin_cmd(card, data, tries_left);
+
+	data->apdu = NULL;
+	return rv;
 }
 
 static struct sc_card_operations cac_ops;
