@@ -799,9 +799,7 @@ static void coolkey_free_private_data(coolkey_private_data_t *priv)
 	list_iterator_stop(l);
 
 	list_destroy(&priv->objects_list);
-	if (priv->token_name) {
-		free(priv->token_name);
-	}
+	free(priv->token_name);
 	free(priv);
 	return;
 }
@@ -1101,6 +1099,8 @@ static int coolkey_read_object(sc_card_t *card, unsigned long object_id, size_t
 	size_t len;
 	int r;
 
+	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
+
 	ulong2bebytes(&params.object_id[0], object_id);
 
 	out_ptr = out_buf;
@@ -1127,7 +1127,7 @@ static int coolkey_read_object(sc_card_t *card, unsigned long object_id, size_t
 	return out_len;
 
 fail:
-	return r;
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 /*
@@ -1208,7 +1208,7 @@ static int coolkey_read_binary(sc_card_t *card, unsigned int idx,
 
 
 	r = coolkey_read_object(card, priv->obj->id, 0, data, priv->obj->length,
-												priv->nonce, sizeof(priv->nonce));
+		priv->nonce, sizeof(priv->nonce));
 	if (r < 0)
 		goto done;
 
@@ -1337,8 +1337,11 @@ static int coolkey_get_token_info(sc_card_t *card, sc_pkcs15_tokeninfo_t * token
 	serial_number = coolkey_cuid_to_string(&priv->cuid);
 
 	if (label && manufacturer_id && serial_number) {
+		free(token_info->label);
 		token_info->label = label;
+		free(token_info->manufacturer_id);
 		token_info->manufacturer_id = manufacturer_id;
+		free(token_info->serial_number);
 		token_info->serial_number = serial_number;
 		return SC_SUCCESS;
 	}
@@ -1367,6 +1370,8 @@ coolkey_fill_object(sc_card_t *card, sc_cardctl_coolkey_object_t *obj)
 	sc_cardctl_coolkey_object_t *obj_entry;
 	coolkey_private_data_t * priv = COOLKEY_DATA(card);
 
+	LOG_FUNC_CALLED(card->ctx);
+
 	if (obj->data != NULL) {
 		return SC_SUCCESS;
 	}
@@ -1378,7 +1383,10 @@ coolkey_fill_object(sc_card_t *card, sc_cardctl_coolkey_object_t *obj)
 				priv->nonce, sizeof(priv->nonce));
 	if (r != (int)buf_len) {
 		free(new_obj_data);
-		return SC_ERROR_CORRUPTED_DATA;
+		if (r < 0) {
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_CORRUPTED_DATA);
 	}
 	obj_entry = coolkey_find_object_by_id(&priv->objects_list, obj->id);
 	if (obj_entry == NULL) {
@@ -1397,7 +1405,7 @@ coolkey_fill_object(sc_card_t *card, sc_cardctl_coolkey_object_t *obj)
 	}
 	obj_entry->data = new_obj_data;
 	obj->data = new_obj_data;
-	return SC_SUCCESS;
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
 /*
@@ -1419,6 +1427,8 @@ coolkey_find_attribute(sc_card_t *card, sc_cardctl_coolkey_attribute_t *attribut
 	attribute->attribute_length = 0;
 	attribute->attribute_value = NULL;
 
+	LOG_FUNC_CALLED(card->ctx);
+
 	if (obj == NULL) {
 		/* cast away const so we can cache the data value */
 		int r = coolkey_fill_object(card, (sc_cardctl_coolkey_object_t *)attribute->object);
@@ -1444,7 +1454,6 @@ coolkey_find_attribute(sc_card_t *card, sc_cardctl_coolkey_attribute_t *attribut
 		return SC_ERROR_CORRUPTED_DATA;
 	}
 
-
 	/*
 	 * now loop through all the attributes in the list. first find the start of the list
 	 */
@@ -1460,7 +1469,7 @@ coolkey_find_attribute(sc_card_t *card, sc_cardctl_coolkey_attribute_t *attribut
 		size_t record_len = coolkey_get_attribute_record_len(attr, object_record_type, buf_len);
 		/* make sure we have the complete record */
 		if (buf_len < record_len || record_len < 4) {
-				return SC_ERROR_CORRUPTED_DATA;
+			return SC_ERROR_CORRUPTED_DATA;
 		}
 		/* does the attribute match the one we are looking for */
 		if (attr_type == coolkey_get_attribute_type(attr, object_record_type, record_len)) {
@@ -1477,7 +1486,7 @@ coolkey_find_attribute(sc_card_t *card, sc_cardctl_coolkey_attribute_t *attribut
 
 		return coolkey_get_attribute_data_fixed(attr_type, fixed_attributes, attribute);
 	}
-	return SC_ERROR_DATA_OBJECT_NOT_FOUND;
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_DATA_OBJECT_NOT_FOUND);
 }
 
 /*
@@ -1664,55 +1673,41 @@ typedef struct coolkey_compute_ecc_params {
 	u8 buf[MAX_COMPUTE_BUF];
 } coolkey_compute_ecc_params_t;
 
-static int coolkey_rsa_op(sc_card_t *card,
-					const u8 * data, size_t datalen,
-					u8 * out, size_t max_out_len)
+static int coolkey_rsa_op(sc_card_t *card, const u8 * data, size_t datalen,
+	u8 * out, size_t max_out_len)
 {
 	int r;
-	const u8 *crypt_in;
-	u8 **crypt_out_p;
-	size_t crypt_in_len, *crypt_out_len_p;
-	coolkey_private_data_t * priv = COOLKEY_DATA(card);
+	u8 **crypt_out_p = NULL;
+	size_t crypt_out_len_p = 0;
+	coolkey_private_data_t *priv = COOLKEY_DATA(card);
 	coolkey_compute_crypt_params_t params;
 	u8 key_number;
 	size_t params_len;
-	size_t buf_len;
-	u8 buf[MAX_COMPUTE_BUF+2];
+	u8 buf[MAX_COMPUTE_BUF + 2];
 	u8 *buf_out;
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
-	sc_log(card->ctx, 
-		 "datalen=%"SC_FORMAT_LEN_SIZE_T"u outlen=%"SC_FORMAT_LEN_SIZE_T"u\n",
-		 datalen, max_out_len);
-
-	crypt_in = data;
-	crypt_in_len = datalen;
-
-	buf_out = &buf[0];
-	crypt_out_p = &buf_out;
-	buf_len = sizeof(buf);
-	crypt_out_len_p = &buf_len;
-	key_number = priv->key_id;
-	params.init.mode = COOLKEY_CRYPT_MODE_RSA_NO_PAD;
-	params.init.location = COOLKEY_CRYPT_LOCATION_APDU;
-	params.init.direction = COOLKEY_CRYPT_DIRECTION_ENCRYPT; /* for no pad, direction is irrelevant */
+	sc_log(card->ctx, "datalen=%"SC_FORMAT_LEN_SIZE_T"u outlen=%"SC_FORMAT_LEN_SIZE_T"u\n",
+		datalen, max_out_len);
 
 	if (priv->key_id > 0xff) {
 		r = SC_ERROR_NO_DEFAULT_KEY;
 		goto done;
 	}
+	key_number = priv->key_id;
 
-	params_len = sizeof(params.init) + crypt_in_len;
+	memset(&params, 0, sizeof(params));
+	params.init.mode = COOLKEY_CRYPT_MODE_RSA_NO_PAD;
+	params.init.direction = COOLKEY_CRYPT_DIRECTION_ENCRYPT; /* for no pad, direction is irrelevant */
 
 	/* send the data to the card if necessary */
-	if (crypt_in_len > MAX_COMPUTE_BUF) {
+	if (datalen > MAX_COMPUTE_BUF) {
+		/* We need to write data to special object on the card as it does not safely fit APDU */
 		u8 len_buf[2];
+
 		params.init.location = COOLKEY_CRYPT_LOCATION_DL_OBJECT;
+
 		params_len = sizeof(params.init);
-		crypt_in = NULL;
-		crypt_in_len = 0;
-		*crypt_out_p = NULL;
-		*crypt_out_len_p = 0;
 
 		ushort2bebytes(len_buf, datalen);
 
@@ -1722,26 +1717,35 @@ static int coolkey_rsa_op(sc_card_t *card,
 			goto done;
 		}
 
-		r = coolkey_write_object(card, COOLKEY_DL_OBJECT_ID, 2, data, datalen, priv->nonce,
-						sizeof(priv->nonce));
+		r = coolkey_write_object(card, COOLKEY_DL_OBJECT_ID, 2, data, datalen, priv->nonce, sizeof(priv->nonce));
 		if (r < 0) {
 			goto done;
 		}
+		ushort2bebytes(params.init.buf_len, 0);
+	} else {
+		/* The data fits in APDU. Copy it to the params object */
+		size_t buf_len;
 
-	}
-	ushort2bebytes(params.init.buf_len, crypt_in_len);
-	if (crypt_in_len) {
-		memcpy(params.buf, crypt_in, crypt_in_len);
-	}
+		params.init.location = COOLKEY_CRYPT_LOCATION_APDU;
 
+		params_len = sizeof(params.init) + datalen;
+
+		buf_out = &buf[0];
+		crypt_out_p = &buf_out;
+		buf_len = sizeof(buf);
+		crypt_out_len_p = buf_len;
+
+		ushort2bebytes(params.init.buf_len, datalen);
+		memcpy(params.buf, data, datalen);
+	}
 
 	r = coolkey_apdu_io(card, COOLKEY_CLASS, COOLKEY_INS_COMPUTE_CRYPT,
 			key_number, COOLKEY_CRYPT_ONE_STEP, (u8 *)&params, params_len,
-			crypt_out_p, crypt_out_len_p, priv->nonce, sizeof(priv->nonce));
-
+			crypt_out_p, &crypt_out_len_p, priv->nonce, sizeof(priv->nonce));
 	if (r < 0) {
 		goto done;
 	}
+
 	if (datalen > MAX_COMPUTE_BUF) {
 		u8 len_buf[2];
 		size_t out_length;
@@ -1760,8 +1764,12 @@ static int coolkey_rsa_op(sc_card_t *card,
 
 	} else {
 		size_t out_length = bebytes2ushort(buf);
+		if (out_length > sizeof buf - 2) {
+			r = SC_ERROR_WRONG_LENGTH;
+			goto done;
+		}
 		out_length = MIN(out_length, max_out_len);
-		memcpy(out, buf+2, out_length);
+		memcpy(out, buf + 2, out_length);
 		r = out_length;
 	}
 
@@ -2047,6 +2055,7 @@ coolkey_process_combined_object(sc_card_t *card, coolkey_private_data_t *priv, u
 	}
 
 	/* store the token name in the priv structure so the emulator can set it */
+	free(priv->token_name);
 	priv->token_name = malloc(decompressed_header->token_name_length+1);
 	if (priv->token_name == NULL) {
 		r = SC_ERROR_OUT_OF_MEMORY;
@@ -2054,17 +2063,28 @@ coolkey_process_combined_object(sc_card_t *card, coolkey_private_data_t *priv, u
 	}
 	memcpy(priv->token_name, &decompressed_header->token_name[0],
 							decompressed_header->token_name_length);
-	priv->token_name[decompressed_header->token_name_length] = 0;
+	priv->token_name[decompressed_header->token_name_length] = '\0';
 	priv->token_name_length = decompressed_header->token_name_length;
 
 
-	for (i=0; i < object_count && object_offset < decompressed_object_len; i++ ) {
-		u8 *current_object = &decompressed_object[object_offset];
-		coolkey_combined_object_header_t *object_header =
-				(coolkey_combined_object_header_t *)current_object;
-		unsigned long object_id = bebytes2ulong(object_header->object_id);
+	for (i=0; i < object_count; i++) {
+		u8 *current_object = NULL;
+		coolkey_combined_object_header_t *object_header = NULL;
+		unsigned long object_id;
 		int current_object_len;
 
+		/* Can we read the object header at all? */
+		if ((object_offset + sizeof(coolkey_combined_object_header_t)) > decompressed_object_len) {
+			r = SC_ERROR_CORRUPTED_DATA;
+			goto done;
+		}
+
+		current_object = &decompressed_object[object_offset];
+		object_header = (coolkey_combined_object_header_t *)current_object;
+
+		/* Parse object ID */
+		object_id = bebytes2ulong(object_header->object_id);
+
 		/* figure out how big it is */
 		r = coolkey_v1_get_object_length(current_object, decompressed_object_len-object_offset);
 		if (r < 0) {
@@ -2078,6 +2098,7 @@ coolkey_process_combined_object(sc_card_t *card, coolkey_private_data_t *priv, u
 		object_offset += current_object_len;
 
 		/* record this object */
+		sc_log(card->ctx, "Add new object id=%ld", object_id);
 		r = coolkey_add_object(priv, object_id, current_object, current_object_len, 1);
 		if (r) {
 			goto done;
@@ -2146,7 +2167,7 @@ static int coolkey_initialize(sc_card_t *card)
 	r = coolkey_list_object(card, COOLKEY_LIST_RESET, &object_info);
 	while (r >= 0) {
 		unsigned long object_id;
-		unsigned short object_len;
+		unsigned long object_len;
 
 		/* The card did not return what we expected: Lets try other objects */
 		if ((size_t)r < (sizeof(object_info)))
@@ -2156,7 +2177,11 @@ static int coolkey_initialize(sc_card_t *card)
 
 		object_id = bebytes2ulong(object_info.object_id);
 		object_len = bebytes2ulong(object_info.object_length);
-
+		/* Avoid insanely large data */
+		if (object_len > MAX_FILE_SIZE) {
+			r = SC_ERROR_CORRUPTED_DATA;
+			goto cleanup;
+		}
 
 		/* the combined object is a single object that can store the other objects.
 		 * most coolkeys provisioned by TPS has a single combined object that is
@@ -2171,7 +2196,7 @@ static int coolkey_initialize(sc_card_t *card)
 				break;
 			}
 			r = coolkey_read_object(card, COOLKEY_COMBINED_OBJECT_ID, 0, object, object_len,
-											priv->nonce, sizeof(priv->nonce));
+				priv->nonce, sizeof(priv->nonce));
 			if (r < 0) {
 				free(object);
 				break;
@@ -2183,6 +2208,7 @@ static int coolkey_initialize(sc_card_t *card)
 			}
 			combined_processed = 1;
 		} else {
+			sc_log(card->ctx, "Add new object id=%ld, len=%lu", object_id, object_len);
 			r = coolkey_add_object(priv, object_id, NULL, object_len, 0);
 			if (r != SC_SUCCESS)
 				sc_log(card->ctx, "coolkey_add_object() returned %d", r);
@@ -2216,19 +2242,19 @@ static int coolkey_initialize(sc_card_t *card)
 		coolkey_make_cuid_from_cplc(&priv->cuid, &cplc_data);
 		priv->token_name = (u8 *)strdup("COOLKEY");
 		if (priv->token_name == NULL) {
-			r= SC_ERROR_OUT_OF_MEMORY;
+			r = SC_ERROR_OUT_OF_MEMORY;
 			goto cleanup;
 		}
 		priv->token_name_length = sizeof("COOLKEY")-1;
 	}
 	card->drv_data = priv;
-	return SC_SUCCESS;
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 
 cleanup:
 	if (priv) {
 		coolkey_free_private_data(priv);
 	}
-	return r;
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 
