@@ -57,13 +57,8 @@
 #ifdef DEBUG_SSL
 static void print_openssl_error(void)
 {
-	static int charge = 0;
 	long r;
 
-	if (!charge) {
-		ERR_load_crypto_strings();
-		charge = 1;
-	}
 	while ((r = ERR_get_error()) != 0)
 		fprintf(stderr, "%s\n", ERR_error_string(r, NULL));
 }
@@ -663,7 +658,10 @@ static int westcos_get_crypte_challenge(sc_card_t * card, const u8 * key,
 {
 	int r;
 #ifdef ENABLE_OPENSSL
-	DES_key_schedule ks1, ks2;
+	EVP_CIPHER_CTX *cctx = NULL;
+	int tmplen = 0;
+	if ((cctx = EVP_CIPHER_CTX_new()) == NULL)
+		return SC_ERROR_INTERNAL;
 #endif
 	u8 buf[8];
 	if ((*len) < sizeof(buf))
@@ -673,9 +671,19 @@ static int westcos_get_crypte_challenge(sc_card_t * card, const u8 * key,
 	if (r)
 		return r;
 #ifdef ENABLE_OPENSSL
-	DES_set_key((const_DES_cblock *) & key[0], &ks1);
-	DES_set_key((const_DES_cblock *) & key[8], &ks2);
-	DES_ecb2_encrypt((const_DES_cblock *)buf, (DES_cblock*)result, &ks1, &ks2, DES_ENCRYPT);
+	if (EVP_EncryptInit_ex(cctx, EVP_des_ede_ecb(), NULL, key, NULL) != 1 ||
+		EVP_CIPHER_CTX_set_padding(cctx,0) != 1 ||
+		EVP_EncryptUpdate(cctx, result, &tmplen, buf, *len) != 1) {
+		EVP_CIPHER_CTX_free(cctx);
+		return SC_ERROR_INTERNAL;
+	}
+	*len = tmplen;
+	if (EVP_EncryptFinal_ex(cctx, result + tmplen, &tmplen)  != 1) {
+		EVP_CIPHER_CTX_free(cctx);
+		return SC_ERROR_INTERNAL;
+    }
+	*len += tmplen;
+	EVP_CIPHER_CTX_free(cctx);
 	return SC_SUCCESS;
 #else
 	return SC_ERROR_NOT_SUPPORTED;
@@ -805,6 +813,7 @@ static int westcos_pin_cmd(sc_card_t * card, struct sc_pin_cmd_data *data,
 			} else {
 				r = SC_ERROR_NOT_SUPPORTED;
 			}
+			data->apdu = NULL;
 		}
 		if (r)
 			return (r);
@@ -1095,15 +1104,17 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 				 const u8 * data, size_t data_len, u8 * out,
 				 size_t outlen)
 {
-	int r;
+	int r = SC_SUCCESS;
 	sc_file_t *keyfile = NULL;
 #ifdef ENABLE_OPENSSL
 	int idx = 0;
 	u8 buf[180];
 	priv_data_t *priv_data = NULL;
 	int pad;
-	RSA *rsa = NULL;
+	EVP_PKEY_CTX *ctx = NULL;
 	BIO *mem = BIO_new(BIO_s_mem());
+	EVP_PKEY *pkey = NULL;
+	size_t tmplen = 0;
 #endif
 
 	if (card == NULL)
@@ -1174,29 +1185,32 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		idx += r;
 	} while (1);
 	BIO_set_mem_eof_return(mem, -1);
-	if (!d2i_RSAPrivateKey_bio(mem, &rsa)) {
+	if (!(pkey = d2i_PrivateKey_bio(mem, NULL))) {
 		sc_log(card->ctx, 
 			"RSA key invalid, %lu\n", ERR_get_error());
 		r = SC_ERROR_UNKNOWN;
 		goto out;
 	}
 
-	/* pkcs11 reset openssl functions */
-	RSA_set_method(rsa, RSA_PKCS1_OpenSSL());
-
-	if ((size_t)RSA_size(rsa) > outlen) {
+	if ((size_t)EVP_PKEY_size(pkey) > outlen) {
 		sc_log(card->ctx,  "Buffer too small\n");
 		r = SC_ERROR_OUT_OF_MEMORY;
 		goto out;
 	}
-#if 1
+
+	if ((ctx = EVP_PKEY_CTX_new(pkey, NULL)) == NULL) {
+		sc_log(card->ctx,  "Can not establish context\n");
+		r = SC_ERROR_UNKNOWN;
+		goto out;
+	}
+
 	if (mode) {		/* decipher */
-		r = RSA_private_decrypt(data_len, data, out, rsa, pad);
-		if (r == -1) {
+		if (EVP_PKEY_decrypt_init(ctx) != 1 ||
+			EVP_PKEY_CTX_set_rsa_padding(ctx, pad) != 1 ||
+			EVP_PKEY_decrypt(ctx, out, &tmplen, data, data_len) != 1) {
 
 #ifdef DEBUG_SSL
 			print_openssl_error();
-
 #endif
 			sc_log(card->ctx, 
 				"Decipher error %lu\n", ERR_get_error());
@@ -1206,13 +1220,12 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 	}
 
 	else {			/* sign */
-
-		r = RSA_private_encrypt(data_len, data, out, rsa, pad);
-		if (r == -1) {
+		if (EVP_PKEY_encrypt_init(ctx) != 1 ||
+			EVP_PKEY_CTX_set_rsa_padding(ctx, pad) != 1 ||
+			EVP_PKEY_encrypt(ctx, out, &tmplen, data, data_len) != 1) {
 
 #ifdef DEBUG_SSL
 			print_openssl_error();
-
 #endif
 			sc_log(card->ctx, 
 				"Signature error %lu\n", ERR_get_error());
@@ -1221,21 +1234,11 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		}
 	}
 
-#else
-	if (RSA_sign(nid, data, data_len, out, &outlen, rsa) != 1) {
-		sc_log(card->ctx, 
-			"RSA_sign error %d \n", ERR_get_error());
-		r = SC_ERROR_UNKNOWN;
-		goto out;
-	}
-	r = outlen;
-
-#endif
 out:
 	if (mem)
 		BIO_free(mem);
-	if (rsa)
-		RSA_free(rsa);
+	EVP_PKEY_free(pkey);
+	EVP_PKEY_CTX_free(ctx);
 out2:
 #endif /* ENABLE_OPENSSL */
 	sc_file_free(keyfile);
