@@ -2,6 +2,7 @@
  * cardctl.h: card_ctl command numbers
  *
  * Copyright (C) 2003  Olaf Kirch <okir@lse.de>
+ * Copyright (C) 2018-2019 GSMK - Gesellschaft für Sichere Mobile Kommunikation mbH
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -62,12 +63,6 @@ enum {
 	SC_CARDCTL_CRYPTOFLEX_BASE = _CTL_PREFIX('C', 'F', 'X'),
 	SC_CARDCTL_CRYPTOFLEX_GENERATE_KEY,
 
-	/*
-	 * MioCOS specific calls
-	 */
-	SC_CARDCTL_MIOCOS_BASE = _CTL_PREFIX('M', 'I', 'O'),
-	SC_CARDCTL_MIOCOS_CREATE_AC,
-
 	/*
 	 * TCOS specific calls
 	 */
@@ -94,12 +89,6 @@ enum {
 	SC_CARDCTL_STARCOS_WRITE_KEY,
 	SC_CARDCTL_STARCOS_GENERATE_KEY,
 
-	/*
-	 * JCOP specific calls
-	 */
-	SC_CARDCTL_JCOP_BASE = _CTL_PREFIX('J', 'C', 'P'),
-	SC_CARDCTL_JCOP_GENERATE_KEY,
-
 	/*
 	 * Oberthur specific calls
 	 */
@@ -263,6 +252,7 @@ enum {
 	SC_CARDCTL_OPENPGP_BASE = _CTL_PREFIX('P', 'G', 'P'),
 	SC_CARDCTL_OPENPGP_GENERATE_KEY,
 	SC_CARDCTL_OPENPGP_STORE_KEY,
+	SC_CARDCTL_OPENPGP_SELECT_DATA,
 
 	/*
 	 * SmartCard-HSM
@@ -273,6 +263,8 @@ enum {
 	SC_CARDCTL_SC_HSM_IMPORT_DKEK_SHARE,
 	SC_CARDCTL_SC_HSM_WRAP_KEY,
 	SC_CARDCTL_SC_HSM_UNWRAP_KEY,
+	SC_CARDCTL_SC_HSM_REGISTER_PUBLIC_KEY,
+	SC_CARDCTL_SC_HSM_PUBLIC_KEY_AUTH_STATUS,
 
 	/*
 	 * DNIe specific calls
@@ -394,26 +386,6 @@ struct sc_cardctl_gpk_genkey {
 	unsigned int		pubkey_len;
 };
 
-enum {
-	SC_CARDCTL_MIOCOS_AC_PIN,
-	SC_CARDCTL_MIOCOS_AC_CHAL,
-	SC_CARDCTL_MIOCOS_AC_LOGICAL,
-	SC_CARDCTL_MIOCOS_AC_SMARTPIN
-};
-
-/*
- * MioCOS AC info
- */
-struct sc_cardctl_miocos_ac_info {
-	int type;
-	int ref;
-	int max_tries;
-	int enable_ac;		/* only applicable to PINs */
-	u8 key_value[8];
-	int max_unblock_tries;	/* same here */
-	u8 unblock_value[8];	/* and here */
-};
-
 /*
  * Siemens CardOS PIN info
  */
@@ -501,14 +473,6 @@ typedef struct sc_starcos_gen_key_data_st {
 	u8	*modulus;
 } sc_starcos_gen_key_data;
 
-struct sc_cardctl_jcop_genkey  {
-     unsigned long exponent;
-     sc_path_t pub_file_ref;
-     sc_path_t pri_file_ref;
-     unsigned char *	pubkey;
-     unsigned int	pubkey_len;
-};
-
 /*
  * Oberthur ex_data stuff
  */
@@ -518,8 +482,6 @@ enum SC_CARDCTL_OBERTHUR_KEY_TYPE {
 	SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC = 0xA1,
 	SC_CARDCTL_OBERTHUR_KEY_RSA_SFM,
 	SC_CARDCTL_OBERTHUR_KEY_RSA_CRT,
-	SC_CARDCTL_OBERTHUR_KEY_DSA_PUBLIC,
-	SC_CARDCTL_OBERTHUR_KEY_DSA_PRIVATE,
 	SC_CARDCTL_OBERTHUR_KEY_EC_CRT,
 	SC_CARDCTL_OBERTHUR_KEY_EC_PUBLIC
 };
@@ -1057,6 +1019,8 @@ typedef struct sc_cardctl_sc_hsm_init_param {
 	struct sc_aid bio2;			/* AID of biometric server for template 2 */
 	u8 options[2];				/* Initialization options */
 	signed char dkek_shares;	/* Number of DKEK shares, 0 for card generated, -1 for none */
+	signed char num_of_pub_keys;         /* Total number of public keys used for public authentication (if > 0) */
+	u8 required_pub_keys;       /* Number of public keys required for authentication (if public auth. is used) */
 	char *label;				/* Token label to be set in EF.TokenInfo (2F03) */
 } sc_cardctl_sc_hsm_init_param_t;
 
@@ -1074,6 +1038,19 @@ typedef struct sc_cardctl_sc_hsm_wrapped_key {
 	size_t wrapped_key_length;	/* Length of key blob */
 } sc_cardctl_sc_hsm_wrapped_key_t;
 
+typedef struct sc_cardctl_sc_hsm_pka_status {
+    u8 num_total;
+    u8 num_missing;
+    u8 num_required;
+    u8 num_authenticated;
+} sc_cardctl_sc_hsm_pka_status_t;
+
+typedef struct sc_cardctl_sc_hsm_pka_register {
+    u8 *buf;
+    size_t buflen;
+    sc_cardctl_sc_hsm_pka_status_t new_status;
+} sc_cardctl_sc_hsm_pka_register_t;
+
 /*
  * isoApplet
  */
