@@ -68,10 +68,6 @@ enum {
 	SC_CARD_TYPE_GPK_GPK8000_16K,
 	SC_CARD_TYPE_GPK_GPK16000 = 3160,
 
-	/* miocos driver */
-	SC_CARD_TYPE_MIOCOS_BASE = 4000,
-	SC_CARD_TYPE_MIOCOS_GENERIC,
-
 	/* mcrd driver */
 	SC_CARD_TYPE_MCRD_BASE = 5000,
 	SC_CARD_TYPE_MCRD_GENERIC,
@@ -94,6 +90,8 @@ enum {
 	SC_CARD_TYPE_STARCOS_GENERIC,
 	SC_CARD_TYPE_STARCOS_V3_4,
 	SC_CARD_TYPE_STARCOS_V3_5,
+	SC_CARD_TYPE_STARCOS_V3_4_ESIGN,
+	SC_CARD_TYPE_STARCOS_V3_5_ESIGN,
 
 	/* tcos driver */
 	SC_CARD_TYPE_TCOS_BASE = 8000,
@@ -108,10 +106,6 @@ enum {
 	SC_CARD_TYPE_OPENPGP_V3,
 	SC_CARD_TYPE_OPENPGP_GNUK,
 
-	/* jcop driver */
-	SC_CARD_TYPE_JCOP_BASE = 10000,
-	SC_CARD_TYPE_JCOP_GENERIC,
-
 	/* oberthur driver */
 	SC_CARD_TYPE_OBERTHUR_BASE = 11000,
 	SC_CARD_TYPE_OBERTHUR_GENERIC,
@@ -273,7 +267,10 @@ enum {
 	SC_CARD_TYPE_IDPRIME_GENERIC,
 
 	/* eDO cards */
-	SC_CARD_TYPE_EDO = 38000
+	SC_CARD_TYPE_EDO = 38000,
+
+	/* JCOP4 cards with NQ-Applet */
+	SC_CARD_TYPE_NQ_APPLET = 39000
 };
 
 extern sc_card_driver_t *sc_get_default_driver(void);
@@ -282,13 +279,11 @@ extern sc_card_driver_t *sc_get_cryptoflex_driver(void);
 extern sc_card_driver_t *sc_get_cyberflex_driver(void);
 extern sc_card_driver_t *sc_get_gpk_driver(void);
 extern sc_card_driver_t *sc_get_gemsafeV1_driver(void);
-extern sc_card_driver_t *sc_get_miocos_driver(void);
 extern sc_card_driver_t *sc_get_mcrd_driver(void);
 extern sc_card_driver_t *sc_get_setcos_driver(void);
 extern sc_card_driver_t *sc_get_starcos_driver(void);
 extern sc_card_driver_t *sc_get_tcos_driver(void);
 extern sc_card_driver_t *sc_get_openpgp_driver(void);
-extern sc_card_driver_t *sc_get_jcop_driver(void);
 extern sc_card_driver_t *sc_get_oberthur_driver(void);
 extern sc_card_driver_t *sc_get_belpic_driver(void);
 extern sc_card_driver_t *sc_get_atrust_acos_driver(void);
@@ -319,6 +314,7 @@ extern sc_card_driver_t *sc_get_npa_driver(void);
 extern sc_card_driver_t *sc_get_esteid2018_driver(void);
 extern sc_card_driver_t *sc_get_idprime_driver(void);
 extern sc_card_driver_t *sc_get_edo_driver(void);
+extern sc_card_driver_t *sc_get_nqApplet_driver(void);
 
 #ifdef __cplusplus
 }
