@@ -933,8 +933,10 @@ iso7816_set_security_env(struct sc_card *card,
 			goto err;
 		}
 	}
-	if (se_num <= 0)
-		return 0;
+	if (se_num <= 0) {
+		r = SC_SUCCESS;
+		goto err;
+	}
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0xF2, se_num);
 	r = sc_transmit_apdu(card, &apdu);
 	sc_unlock(card);
@@ -1165,8 +1167,8 @@ iso7816_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries_l
 
 	/* Many cards do support PIN status queries, but some cards don't and
 	 * mistakenly count the command as a failed PIN attempt, so for now we
-	 * whitelist cards with this flag.  In future this may be reduced to a
-	 * blacklist, subject to testing more cards. */
+	 * allow cards with this flag.  In future this may be reduced to a
+	 * blocklist, subject to testing more cards. */
 	if (data->cmd == SC_PIN_CMD_GET_INFO &&
 	    !(card->caps & SC_CARD_CAP_ISO7816_PIN_INFO)) {
 		sc_log(card->ctx, "Card does not support PIN status queries");
@@ -1324,7 +1326,9 @@ static struct sc_card_operations iso_ops = {
 	NULL,			/* read_public_key */
 	NULL,			/* card_reader_lock_obtained */
 	NULL,			/* wrap */
-	NULL			/* unwrap */
+	NULL,			/* unwrap */
+	NULL,			/* encrypt_sym */
+	NULL			/* decrypt_sym */
 };
 
 static struct sc_card_driver iso_driver = {
@@ -1378,7 +1382,7 @@ int iso7816_read_binary_sfid(sc_card_t *card, unsigned char sfid,
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r < 0 && r != SC_ERROR_FILE_END_REACHED)
 		goto err;
-	/* emulate the behaviour of sc_read_binary */
+	/* emulate the behaviour of iso7816_read_binary */
 	r = apdu.resplen;
 
 	while(1) {
@@ -1386,11 +1390,13 @@ int iso7816_read_binary_sfid(sc_card_t *card, unsigned char sfid,
 			*ef_len += r;
 			break;
 		}
-		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
-			break;
-		if (r < 0) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_VERBOSE, "Could not read EF.");
-			goto err;
+		if (r <= 0) {
+			if (*ef_len > 0)
+				break;
+			else {
+				sc_debug(card->ctx, SC_LOG_DEBUG_VERBOSE, "Could not read EF.");
+				goto err;
+			}
 		}
 		*ef_len += r;
 
@@ -1401,7 +1407,7 @@ int iso7816_read_binary_sfid(sc_card_t *card, unsigned char sfid,
 		}
 		*ef = p;
 
-		r = sc_read_binary(card, *ef_len,
+		r = iso7816_read_binary(card, *ef_len,
 				*ef + *ef_len, read, 0);
 	}
 
