@@ -59,9 +59,15 @@ enum {
 #define __FUNCTION__ NULL
 #endif
 
+#ifdef __FILE_NAME__
+#define FILENAME __FILE_NAME__
+#else
+#define FILENAME __FILE__
+#endif
+
 #if defined(__GNUC__)
-#define sc_debug(ctx, level, format, args...)	sc_do_log(ctx, level, __FILE__, __LINE__, __FUNCTION__, format , ## args)
-#define sc_log(ctx, format, args...)   sc_do_log(ctx, SC_LOG_DEBUG_NORMAL, __FILE__, __LINE__, __FUNCTION__, format , ## args)
+#define sc_debug(ctx, level, format, args...)	sc_do_log(ctx, level, FILENAME, __LINE__, __FUNCTION__, format , ## args)
+#define sc_log(ctx, format, args...)   sc_do_log(ctx, SC_LOG_DEBUG_NORMAL, FILENAME, __LINE__, __FUNCTION__, format , ## args)
 #else
 #define sc_debug _sc_debug
 #define sc_log _sc_log
@@ -109,7 +115,7 @@ int sc_color_fprintf(int colors, struct sc_context *ctx, FILE * stream, const ch
  * @param[in] len   Length of \a data
  */
 #define sc_debug_hex(ctx, level, label, data, len) \
-    _sc_debug_hex(ctx, level, __FILE__, __LINE__, __FUNCTION__, label, data, len)
+    _sc_debug_hex(ctx, level, FILENAME, __LINE__, __FUNCTION__, label, data, len)
 #define sc_log_hex(ctx, label, data, len) \
     sc_debug_hex(ctx, SC_LOG_DEBUG_NORMAL, label, data, len)
 /** 
@@ -131,17 +137,17 @@ void sc_hex_dump(const u8 *buf, size_t len, char *out, size_t outlen);
 const char * sc_dump_hex(const u8 * in, size_t count);
 const char * sc_dump_oid(const struct sc_object_id *oid);
 #define SC_FUNC_CALLED(ctx, level) do { \
-	 sc_do_log(ctx, level, __FILE__, __LINE__, __FUNCTION__, "called\n"); \
+	 sc_do_log(ctx, level, FILENAME, __LINE__, __FUNCTION__, "called\n"); \
 } while (0)
 #define LOG_FUNC_CALLED(ctx) SC_FUNC_CALLED((ctx), SC_LOG_DEBUG_NORMAL)
 
 #define SC_FUNC_RETURN(ctx, level, r) do { \
 	int _ret = r; \
 	if (_ret <= 0) { \
-		sc_do_log_color(ctx, level, __FILE__, __LINE__, __FUNCTION__, _ret ? SC_COLOR_FG_RED : 0, \
+		sc_do_log_color(ctx, level, FILENAME, __LINE__, __FUNCTION__, _ret ? SC_COLOR_FG_RED : 0, \
 			"returning with: %d (%s)\n", _ret, sc_strerror(_ret)); \
 	} else { \
-		sc_do_log(ctx, level, __FILE__, __LINE__, __FUNCTION__, \
+		sc_do_log(ctx, level, FILENAME, __LINE__, __FUNCTION__, \
 			"returning with: %d\n", _ret); \
 	} \
 	return _ret; \
@@ -151,7 +157,7 @@ const char * sc_dump_oid(const struct sc_object_id *oid);
 #define SC_TEST_RET(ctx, level, r, text) do { \
 	int _ret = (r); \
 	if (_ret < 0) { \
-		sc_do_log_color(ctx, level, __FILE__, __LINE__, __FUNCTION__, SC_COLOR_FG_RED, \
+		sc_do_log_color(ctx, level, FILENAME, __LINE__, __FUNCTION__, SC_COLOR_FG_RED, \
 			"%s: %d (%s)\n", (text), _ret, sc_strerror(_ret)); \
 		return _ret; \
 	} \
@@ -161,7 +167,7 @@ const char * sc_dump_oid(const struct sc_object_id *oid);
 #define SC_TEST_GOTO_ERR(ctx, level, r, text) do { \
 	int _ret = (r); \
 	if (_ret < 0) { \
-		sc_do_log_color(ctx, level, __FILE__, __LINE__, __FUNCTION__, SC_COLOR_FG_RED, \
+		sc_do_log_color(ctx, level, FILENAME, __LINE__, __FUNCTION__, SC_COLOR_FG_RED, \
 			"%s: %d (%s)\n", (text), _ret, sc_strerror(_ret)); \
 		goto err; \
 	} \
