@@ -26,6 +26,7 @@
 #include <string.h>
 #include <stdio.h>
 
+#include "internal.h"
 #include "common/compat_strlcpy.h"
 #include "pkcs15.h"
 #include "cardctl.h"
@@ -167,17 +168,14 @@ static int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
 	r = sc_bin_to_hex(serial.value, serial.len, buf, sizeof(buf), 0);
 	if (r != SC_SUCCESS)
 		return SC_ERROR_INTERNAL;
-	if (p15card->tokeninfo->serial_number)
-		free(p15card->tokeninfo->serial_number);
-	p15card->tokeninfo->serial_number = strdup(buf);
+
+	set_string(&p15card->tokeninfo->serial_number, buf);
 	if (!p15card->tokeninfo->serial_number)
 		return SC_ERROR_INTERNAL;
 	/* the manufacturer ID, in this case Giesecke & Devrient GmbH */
-	if (p15card->tokeninfo->manufacturer_id)
-		free(p15card->tokeninfo->manufacturer_id);
-	p15card->tokeninfo->manufacturer_id = strdup(MANU_ID);
+	set_string(&p15card->tokeninfo->manufacturer_id, MANU_ID);
 	if (!p15card->tokeninfo->manufacturer_id)
-		return SC_ERROR_INTERNAL;
+		goto err;
 
 	/* set certs */
 	for (i = 0; certs[i].label; i++) {
@@ -199,7 +197,7 @@ static int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
 
 		r = sc_pkcs15emu_add_x509_cert(p15card, &cert_obj, &cert_info);
 		if (r < 0)
-			return SC_ERROR_INTERNAL;
+			goto err;
 	}
 	/* set pins */
 	for (i = 0; pins[i].label; i++) {
@@ -228,7 +226,7 @@ static int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
 
 		r = sc_pkcs15emu_add_pin_obj(p15card, &pin_obj, &pin_info);
 		if (r < 0)
-			return SC_ERROR_INTERNAL;
+			goto err;
 	}
 	/* set private keys */
 	for (i = 0; prkeys[i].label; i++) {
@@ -252,20 +250,23 @@ static int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
 
 		r = sc_pkcs15emu_add_rsa_prkey(p15card, &prkey_obj, &prkey_info);
 		if (r < 0)
-			return SC_ERROR_INTERNAL;
+			goto err;
 	}
 		
 	/* select the application DF */
 	sc_format_path("3F00DF01", &path);
 	r = sc_select_file(card, &path, &file);
 	if (r != SC_SUCCESS || !file)
-		return SC_ERROR_INTERNAL;
+		goto err;
 	/* set the application DF */
-	if (p15card->file_app)
-		free(p15card->file_app);
+	sc_file_free(p15card->file_app);
 	p15card->file_app = file;
 
 	return SC_SUCCESS;
+
+err:
+	sc_pkcs15_card_clear(p15card);
+	return SC_ERROR_INTERNAL;
 }
 
 int sc_pkcs15emu_starcert_init_ex(sc_pkcs15_card_t *p15card,
