@@ -33,9 +33,9 @@
 #include "asn1.h"
 #include "pkcs15.h"
 #include "pkcs15-syn.h"
+#include "pkcs15-emulator-filter.h"
 
 struct sc_pkcs15_emulator_handler builtin_emulators[] = {
-	{ "westcos",	sc_pkcs15emu_westcos_init_ex	},
 	{ "openpgp",	sc_pkcs15emu_openpgp_init_ex	},
 	{ "starcert",	sc_pkcs15emu_starcert_init_ex	},
 	{ "tcos",	sc_pkcs15emu_tcos_init_ex	},
@@ -44,11 +44,7 @@ struct sc_pkcs15_emulator_handler builtin_emulators[] = {
 	{ "PIV-II",     sc_pkcs15emu_piv_init_ex	},
 	{ "cac",        sc_pkcs15emu_cac_init_ex	},
 	{ "idprime",    sc_pkcs15emu_idprime_init_ex	},
-	{ "gemsafeGPK",	sc_pkcs15emu_gemsafeGPK_init_ex	},
 	{ "gemsafeV1",	sc_pkcs15emu_gemsafeV1_init_ex	},
-	{ "actalis",	sc_pkcs15emu_actalis_init_ex	},
-	{ "atrust-acos",sc_pkcs15emu_atrust_acos_init_ex},
-	{ "tccardos",	sc_pkcs15emu_tccardos_init_ex	},
 	{ "entersafe",  sc_pkcs15emu_entersafe_init_ex	},
 	{ "pteid",	sc_pkcs15emu_pteid_init_ex	},
 	{ "oberthur",   sc_pkcs15emu_oberthur_init_ex	},
@@ -61,10 +57,20 @@ struct sc_pkcs15_emulator_handler builtin_emulators[] = {
 	{ "din66291",   sc_pkcs15emu_din_66291_init_ex	},
 	{ "esteid2018", sc_pkcs15emu_esteid2018_init_ex	},
 	{ "cardos",     sc_pkcs15emu_cardos_init_ex	},
-
+	{ "nqapplet",   sc_pkcs15emu_nqapplet_init_ex },
+	{ "esign",      sc_pkcs15emu_starcos_esign_init_ex },
 	{ NULL, NULL }
 };
 
+struct sc_pkcs15_emulator_handler old_emulators[] = {
+	{ "westcos",	sc_pkcs15emu_westcos_init_ex	},
+	{ "gemsafeGPK",	sc_pkcs15emu_gemsafeGPK_init_ex	},
+	{ "atrust-acos",sc_pkcs15emu_atrust_acos_init_ex},
+	{ "actalis",	sc_pkcs15emu_actalis_init_ex	},
+	{ "tccardos",	sc_pkcs15emu_tccardos_init_ex	},
+	{ NULL, NULL }
+};	
+
 static int parse_emu_block(sc_pkcs15_card_t *, struct sc_aid *, scconf_block *);
 static sc_pkcs15_df_t * sc_pkcs15emu_get_df(sc_pkcs15_card_t *p15card,
 	unsigned int type);
@@ -91,6 +97,8 @@ int sc_pkcs15_is_emulation_only(sc_card_t *card)
 		case SC_CARD_TYPE_DNIE_USER:
 		case SC_CARD_TYPE_DNIE_TERMINATED:
 		case SC_CARD_TYPE_IASECC_GEMALTO:
+		case SC_CARD_TYPE_IASECC_CPX:
+		case SC_CARD_TYPE_IASECC_CPXCL:
 		case SC_CARD_TYPE_PIV_II_GENERIC:
 		case SC_CARD_TYPE_PIV_II_HIST:
 		case SC_CARD_TYPE_PIV_II_NEO:
@@ -98,6 +106,9 @@ int sc_pkcs15_is_emulation_only(sc_card_t *card)
 		case SC_CARD_TYPE_ESTEID_2018:
 		case SC_CARD_TYPE_CARDOS_V5_0:
 		case SC_CARD_TYPE_CARDOS_V5_3:
+		case SC_CARD_TYPE_NQ_APPLET:
+		case SC_CARD_TYPE_STARCOS_V3_4_ESIGN:
+		case SC_CARD_TYPE_STARCOS_V3_5_ESIGN:
 
 			return 1;
 		default:
@@ -130,25 +141,34 @@ sc_pkcs15_bind_synthetic(sc_pkcs15_card_t *p15card, struct sc_aid *aid)
 	} else {
 		/* we have a conf file => let's use it */
 		int builtin_enabled;
-		const scconf_list *list, *item;
+		const scconf_list *list;
 
 		builtin_enabled = scconf_get_bool(conf_block, "enable_builtin_emulation", 1);
 		list = scconf_find_list(conf_block, "builtin_emulators"); /* FIXME: rename to enabled_emulators */
 
 		if (builtin_enabled && list) {
-			/* get the list of enabled emulation drivers */
-			for (item = list; item; item = item->next) {
-				/* go through the list of builtin drivers */
-				const char *name = item->data;
+			/* filter enabled emulation drivers from conf file */
+			struct _sc_pkcs15_emulators filtered_emulators;
+			struct sc_pkcs15_emulator_handler** lst;
+			int ret;
 
-				sc_log(ctx, "trying %s", name);
-				for (i = 0; builtin_emulators[i].name; i++)
-					if (!strcmp(builtin_emulators[i].name, name)) {
-						r = builtin_emulators[i].handler(p15card, aid);
-						if (r == SC_SUCCESS)
-							/* we got a hit */
-							goto out;
-					}
+			filtered_emulators.ccount = 0;
+			ret = set_emulators(ctx, &filtered_emulators, list, builtin_emulators, old_emulators);
+			if (ret == SC_SUCCESS || ret == SC_ERROR_TOO_MANY_OBJECTS) {
+				lst = filtered_emulators.list_of_handlers;
+
+				if (ret == SC_ERROR_TOO_MANY_OBJECTS)
+					sc_log(ctx, "trying first %d emulators from conf file", SC_MAX_PKCS15_EMULATORS);
+
+				for (i = 0; lst[i]; i++) {
+					sc_log(ctx, "trying %s", lst[i]->name);
+					r = lst[i]->handler(p15card, aid);
+					if (r == SC_SUCCESS)
+						/* we got a hit */
+						goto out;
+				}
+			} else {
+				sc_log(ctx, "failed to filter enabled card emulators: %s", sc_strerror(ret));
 			}
 		}
 		else if (builtin_enabled) {
@@ -220,6 +240,14 @@ static int parse_emu_block(sc_pkcs15_card_t *p15card, struct sc_aid *aid, scconf
 				break;
 			}
 		}
+		if (init_func_ex == NULL) {
+			for (i = 0; old_emulators[i].name; i++) {
+				if (!strcmp(old_emulators[i].name, module_name)) {
+					init_func_ex = old_emulators[i].handler;
+					break;
+				}
+			}
+		}
 	} else {
 		const char *(*get_version)(void);
 		const char *name = NULL;
@@ -364,6 +392,7 @@ int sc_pkcs15emu_add_ec_prkey(sc_pkcs15_card_t *p15card,
 
 	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PRKEY_EC, obj, &key);
 }
+
 int sc_pkcs15emu_add_ec_pubkey(sc_pkcs15_card_t *p15card,
 	const sc_pkcs15_object_t *obj, const sc_pkcs15_pubkey_info_t *in_key)
 {
@@ -375,6 +404,56 @@ int sc_pkcs15emu_add_ec_pubkey(sc_pkcs15_card_t *p15card,
 	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PUBKEY_EC, obj, &key);
 }
 
+int sc_pkcs15emu_add_eddsa_prkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_prkey_info_t *in_key)
+{
+	sc_pkcs15_prkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_SENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
+				| SC_PKCS15_PRKEY_ACCESS_LOCAL;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PRKEY_EDDSA, obj, &key);
+}
+
+int sc_pkcs15emu_add_eddsa_pubkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_pubkey_info_t *in_key)
+{
+	sc_pkcs15_pubkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PUBKEY_EDDSA, obj, &key);
+}
+
+int sc_pkcs15emu_add_xeddsa_prkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_prkey_info_t *in_key)
+{
+	sc_pkcs15_prkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_SENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
+				| SC_PKCS15_PRKEY_ACCESS_NEVEREXTRACTABLE
+				| SC_PKCS15_PRKEY_ACCESS_LOCAL;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PRKEY_XEDDSA, obj, &key);
+}
+
+int sc_pkcs15emu_add_xeddsa_pubkey(sc_pkcs15_card_t *p15card,
+	const sc_pkcs15_object_t *obj, const sc_pkcs15_pubkey_info_t *in_key)
+{
+	sc_pkcs15_pubkey_info_t key = *in_key;
+
+	if (key.access_flags == 0)
+		key.access_flags = SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE;
+
+	return sc_pkcs15emu_object_add(p15card, SC_PKCS15_TYPE_PUBKEY_XEDDSA, obj, &key);
+}
+
 int sc_pkcs15emu_add_x509_cert(sc_pkcs15_card_t *p15card,
 	const sc_pkcs15_object_t *obj, const sc_pkcs15_cert_info_t *cert)
 {
