@@ -182,8 +182,10 @@ static int pcsc_to_opensc_error(LONG rv)
 		return SC_ERROR_CARD_UNRESPONSIVE;
 	case SCARD_E_SHARING_VIOLATION:
 		return SC_ERROR_READER_LOCKED;
+#ifdef SCARD_E_NO_READERS_AVAILABLE
 	case SCARD_E_NO_READERS_AVAILABLE:
 		return SC_ERROR_NO_READERS_FOUND;
+#endif
 	case SCARD_E_UNKNOWN_READER:
 		return SC_ERROR_READER_DETACHED;
 
@@ -391,10 +393,7 @@ static int refresh_attributes(sc_reader_t *reader)
 				|| rv == (LONG)SCARD_E_NO_READERS_AVAILABLE
 #endif
 				|| rv == (LONG)SCARD_E_SERVICE_STOPPED) {
- 			if (old_flags & SC_READER_CARD_PRESENT) {
- 				reader->flags |= SC_READER_CARD_CHANGED;
- 			}
-			
+ 			reader->flags &= ~(SC_READER_CARD_PRESENT);
  			SC_FUNC_RETURN(reader->ctx, SC_LOG_DEBUG_VERBOSE, SC_SUCCESS);
  		}
 
@@ -413,7 +412,8 @@ static int refresh_attributes(sc_reader_t *reader)
 		 * XXX: We'll hit it again, as no readers are removed currently.
 		 */
 		reader->flags &= ~(SC_READER_CARD_PRESENT);
-		return SC_ERROR_READER_DETACHED;
+		sc_log(reader->ctx, "Reader unknown: %s", sc_strerror(SC_ERROR_READER_DETACHED));
+		SC_FUNC_RETURN(reader->ctx, SC_LOG_DEBUG_VERBOSE, SC_SUCCESS);
 	}
 
 	reader->flags &= ~(SC_READER_CARD_CHANGED|SC_READER_CARD_INUSE|SC_READER_CARD_EXCLUSIVE);
@@ -456,8 +456,6 @@ static int refresh_attributes(sc_reader_t *reader)
 		}
 	} else {
 		reader->flags &= ~SC_READER_CARD_PRESENT;
-		if (old_flags & SC_READER_CARD_PRESENT)
-			reader->flags |= SC_READER_CARD_CHANGED;
 	}
 	sc_log(reader->ctx, "card %s%s",
 			reader->flags & SC_READER_CARD_PRESENT ? "present" : "absent",
@@ -474,7 +472,12 @@ static int pcsc_detect_card_presence(sc_reader_t *reader)
 	rv = refresh_attributes(reader);
 	if (rv != SC_SUCCESS)
 		LOG_FUNC_RETURN(reader->ctx, rv);
-	LOG_FUNC_RETURN(reader->ctx, reader->flags);
+
+	// Return 0 if the card is not present
+	if (reader->flags & SC_READER_CARD_PRESENT)
+		LOG_FUNC_RETURN(reader->ctx, reader->flags);
+	else
+		LOG_FUNC_RETURN(reader->ctx, 0);
 }
 
 static int check_forced_protocol(sc_reader_t *reader, DWORD *protocol)
@@ -583,7 +586,8 @@ static void initialize_uid(sc_reader_t *reader)
 		apdu.resplen = sizeof rbuf;
 
 		if (SC_SUCCESS == pcsc_transmit(reader, &apdu)
-				&& apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
+				&& apdu.sw1 == 0x90 && apdu.sw2 == 0x00
+				&& 0 < apdu.resplen && apdu.resplen <= SC_MAX_UID_SIZE) {
 			reader->uid.len = apdu.resplen;
 			memcpy(reader->uid.value, apdu.resp, reader->uid.len);
 			sc_log_hex(reader->ctx, "UID",
@@ -668,7 +672,7 @@ static int pcsc_disconnect(sc_reader_t * reader)
 		LONG rv = priv->gpriv->SCardDisconnect(priv->pcsc_card, priv->gpriv->disconnect_action);
 		PCSC_TRACE(reader, "SCardDisconnect returned", rv);
 	}
-	reader->flags = 0;
+	reader->flags &= SC_READER_REMOVED;
 	return SC_SUCCESS;
 }
 
@@ -1131,9 +1135,8 @@ static void detect_reader_features(sc_reader_t *reader, SCARDHANDLE card_handle)
 	sc_context_t *ctx = reader->ctx;
 	struct pcsc_global_private_data *gpriv = (struct pcsc_global_private_data *) ctx->reader_drv_data;
 	struct pcsc_private_data *priv = reader->drv_data;
-	u8 feature_buf[256], rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	DWORD rcount, feature_len, i;
-	PCSC_TLV_STRUCTURE *pcsc_tlv;
+	DWORD rcount, i;
+	u8 buf[256];
 	LONG rv;
 	const char *log_disabled = "but it's disabled in configuration file";
 	int id_vendor = 0, id_product = 0;
@@ -1145,43 +1148,41 @@ static void detect_reader_features(sc_reader_t *reader, SCARDHANDLE card_handle)
 	if (gpriv->SCardControl == NULL)
 		return;
 
-	rv = gpriv->SCardControl(card_handle, CM_IOCTL_GET_FEATURE_REQUEST, NULL, 0, feature_buf, sizeof(feature_buf), &feature_len);
+	rv = gpriv->SCardControl(card_handle, CM_IOCTL_GET_FEATURE_REQUEST, NULL, 0, buf, sizeof(buf), &rcount);
 	if (rv != SCARD_S_SUCCESS) {
 		PCSC_TRACE(reader, "SCardControl failed", rv);
 		return;
 	}
 
-	if ((feature_len % sizeof(PCSC_TLV_STRUCTURE)) != 0) {
+	if ((rcount % sizeof(PCSC_TLV_STRUCTURE)) != 0
+			|| rcount > sizeof buf) {
 		sc_log(ctx, "Inconsistent TLV from reader!");
 		return;
 	}
 
-	/* get the number of elements instead of the complete size */
-	feature_len /= sizeof(PCSC_TLV_STRUCTURE);
-
-	pcsc_tlv = (PCSC_TLV_STRUCTURE *)feature_buf;
-	for (i = 0; i < feature_len; i++) {
-		sc_log(ctx, "Reader feature %02x found", pcsc_tlv[i].tag);
-		if (pcsc_tlv[i].tag == FEATURE_VERIFY_PIN_DIRECT) {
-			priv->verify_ioctl = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_VERIFY_PIN_START) {
-			priv->verify_ioctl_start = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_VERIFY_PIN_FINISH) {
-			priv->verify_ioctl_finish = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_MODIFY_PIN_DIRECT) {
-			priv->modify_ioctl = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_MODIFY_PIN_START) {
-			priv->modify_ioctl_start = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_MODIFY_PIN_FINISH) {
-			priv->modify_ioctl_finish = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_IFD_PIN_PROPERTIES) {
-			priv->pin_properties_ioctl = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_GET_TLV_PROPERTIES)  {
-			priv->get_tlv_properties = ntohl(pcsc_tlv[i].value);
-		} else if (pcsc_tlv[i].tag == FEATURE_EXECUTE_PACE) {
-			priv->pace_ioctl = ntohl(pcsc_tlv[i].value);
+	for (i = 0; i < rcount; i += sizeof(PCSC_TLV_STRUCTURE)) {
+		PCSC_TLV_STRUCTURE *pcsc_tlv = (PCSC_TLV_STRUCTURE *)(buf+i);
+		sc_log(ctx, "Reader feature %02x found", pcsc_tlv->tag);
+		if (pcsc_tlv->tag == FEATURE_VERIFY_PIN_DIRECT) {
+			priv->verify_ioctl = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_VERIFY_PIN_START) {
+			priv->verify_ioctl_start = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_VERIFY_PIN_FINISH) {
+			priv->verify_ioctl_finish = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_MODIFY_PIN_DIRECT) {
+			priv->modify_ioctl = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_MODIFY_PIN_START) {
+			priv->modify_ioctl_start = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_MODIFY_PIN_FINISH) {
+			priv->modify_ioctl_finish = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_IFD_PIN_PROPERTIES) {
+			priv->pin_properties_ioctl = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_GET_TLV_PROPERTIES)  {
+			priv->get_tlv_properties = ntohl(pcsc_tlv->value);
+		} else if (pcsc_tlv->tag == FEATURE_EXECUTE_PACE) {
+			priv->pace_ioctl = ntohl(pcsc_tlv->value);
 		} else {
-			sc_log(ctx, "Reader feature %02x is not supported", pcsc_tlv[i].tag);
+			sc_log(ctx, "Reader feature %02x is not supported", pcsc_tlv->tag);
 		}
 	}
 
@@ -1219,12 +1220,13 @@ static void detect_reader_features(sc_reader_t *reader, SCARDHANDLE card_handle)
 
 	/* Detect display */
 	if (priv->pin_properties_ioctl) {
-		rcount = sizeof(rbuf);
-		rv = gpriv->SCardControl(card_handle, priv->pin_properties_ioctl, NULL, 0, rbuf, sizeof(rbuf), &rcount);
+		rcount = sizeof(buf);
+		rv = gpriv->SCardControl(card_handle, priv->pin_properties_ioctl,
+			NULL, 0, buf, sizeof(buf), &rcount);
 		if (rv == SCARD_S_SUCCESS) {
 #ifdef PIN_PROPERTIES_v5
 			if (rcount == sizeof(PIN_PROPERTIES_STRUCTURE_v5)) {
-				PIN_PROPERTIES_STRUCTURE_v5 *caps = (PIN_PROPERTIES_STRUCTURE_v5 *)rbuf;
+				PIN_PROPERTIES_STRUCTURE_v5 *caps = (PIN_PROPERTIES_STRUCTURE_v5 *) &buf;
 				if (caps->wLcdLayout > 0) {
 					sc_log(ctx, "Reader has a display: %04X", caps->wLcdLayout);
 					reader->capabilities |= SC_READER_CAP_DISPLAY;
@@ -1233,7 +1235,7 @@ static void detect_reader_features(sc_reader_t *reader, SCARDHANDLE card_handle)
 			}
 #endif
 			if (rcount == sizeof(PIN_PROPERTIES_STRUCTURE)) {
-				PIN_PROPERTIES_STRUCTURE *caps = (PIN_PROPERTIES_STRUCTURE *)rbuf;
+				PIN_PROPERTIES_STRUCTURE *caps = (PIN_PROPERTIES_STRUCTURE *) buf;
 				if (caps->wLcdLayout > 0) {
 					sc_log(ctx, "Reader has a display: %04X", caps->wLcdLayout);
 					reader->capabilities |= SC_READER_CAP_DISPLAY;
@@ -1292,13 +1294,13 @@ static void detect_reader_features(sc_reader_t *reader, SCARDHANDLE card_handle)
 	}
 
 	if(gpriv->SCardGetAttrib != NULL) {
-		rcount = sizeof(rbuf);
+		rcount = sizeof(buf);
 		if (gpriv->SCardGetAttrib(card_handle, SCARD_ATTR_VENDOR_NAME,
-					rbuf, &rcount) == SCARD_S_SUCCESS
+					buf, &rcount) == SCARD_S_SUCCESS
 				&& rcount > 0) {
 			/* add NUL termination, just in case... */
-			rbuf[(sizeof rbuf)-1] = '\0';
-			reader->vendor = strdup((char *) rbuf);
+			buf[(sizeof buf)-1] = '\0';
+			reader->vendor = strdup((char *) buf);
 		}
 
 		rcount = sizeof i;
@@ -1789,6 +1791,10 @@ out:
 				/* no other event has been detected, yet */
 				*event_reader = gpriv->attached_reader;
 				*event = SC_EVENT_READER_ATTACHED;
+				/* If the card is present in the reader, report also this event */
+				if (gpriv->attached_reader->flags & SC_READER_CARD_PRESENT) {
+					*event |= SC_EVENT_CARD_INSERTED;
+				}
 				r = SC_SUCCESS;
 			}
 			gpriv->attached_reader = NULL;
@@ -2496,6 +2502,31 @@ static void detect_protocol(sc_reader_t *reader, SCARDHANDLE card_handle)
 	reader->active_protocol = pcsc_proto_to_opensc(prot);
 }
 
+int
+pcsc_check_reader_handles(sc_context_t *ctx, sc_reader_t *reader, void * pcsc_context_handle, void * pcsc_card_handle)
+{
+	char reader_name[128];
+	DWORD reader_name_size = sizeof(reader_name);
+
+	if (NULL == reader)
+		return 1;
+
+	struct pcsc_private_data *priv = reader->drv_data;
+	memset(reader_name, 0, sizeof(reader_name));
+
+	/* check if new handles are for the same reader as old handles */
+	if (SCARD_S_SUCCESS != priv->gpriv->SCardGetAttrib(*(SCARDHANDLE *)pcsc_card_handle,
+				SCARD_ATTR_DEVICE_SYSTEM_NAME_A, (LPBYTE)
+				reader_name, &reader_name_size)
+			|| strcmp(reader_name, reader->name) != 0) {
+		sc_log(ctx, "Reader name changed from \"%s\" to \"%s\"", reader->name, reader_name);
+
+		return 1;
+	}
+
+	return 0;
+}
+
 int pcsc_use_reader(sc_context_t *ctx, void * pcsc_context_handle, void * pcsc_card_handle)
 {
 	SCARDHANDLE card_handle;
