@@ -49,19 +49,20 @@ sc_simpletlv_put_tag(u8 tag, size_t datalen, u8 *out, size_t outlen, u8 **ptr)
 	/* tag is just number between 0x01 and 0xFE */
 	if (tag == 0x00 || tag == 0xff)
 		return SC_ERROR_INVALID_ARGUMENTS;
+	if (datalen > 0xffff) {
+		/* we can't store more than two bytes in Simple TLV */
+		return SC_ERROR_WRONG_LENGTH;
+	}
 
 	*p++ = tag; /* tag is single byte */
 	if (datalen < 0xff) {
 		/* short value up to 255 */
 		*p++ = (u8)datalen; /* is in the second byte */
-	} else if (datalen < 0xffff) {
+	} else {
 		/* longer values up to 65535 */
 		*p++ = (u8)0xff; /* first byte is 0xff */
 		*p++ = (u8)datalen & 0xff;
 		*p++ = (u8)(datalen >> 8) & 0xff; /* LE */
-	} else {
-		/* we can't store more than two bytes in Simple TLV */
-		return SC_ERROR_WRONG_LENGTH;
 	}
 	if (ptr != NULL)
 		*ptr = p;
@@ -72,29 +73,40 @@ sc_simpletlv_put_tag(u8 tag, size_t datalen, u8 *out, size_t outlen, u8 **ptr)
  * content.
  */
 int
-sc_simpletlv_read_tag(u8 **buf, size_t buflen, u8 *tag_out, size_t *taglen)
+sc_simpletlv_read_tag(const u8 **buf, size_t buflen, u8 *tag_out, size_t *taglen)
 {
-	size_t len;
-	u8 *p = *buf;
+	u8 tag;
+	size_t left = buflen, len;
+	const u8 *p = *buf;
 
-	if (buflen < 2) {
-		*buf = p+buflen;
-		return SC_ERROR_INVALID_ARGUMENTS;
+	*buf = NULL;
+
+	if (left < 2) {
+		return SC_ERROR_INVALID_TLV_OBJECT;
 	}
+	tag = *p;
+	p++;
+	len = *p;
+	p++;
+	left -= 2;
 
-	*tag_out = *p++;
-	len = *p++;
 	if (len == 0xff) {
 		/* don't crash on bad data */
-		if (buflen < 4) {
-			*taglen = 0;
-			return SC_ERROR_INVALID_ARGUMENTS;
+		if (left < 2) {
+			return SC_ERROR_INVALID_TLV_OBJECT;
 		}
 		/* skip two bytes (the size) */
 		len = lebytes2ushort(p);
-		p+=2;
+		p += 2;
+		left -= 2;
 	}
+
+	*tag_out = tag;
 	*taglen = len;
 	*buf = p;
+
+	if (len > left)
+		return SC_ERROR_TLV_END_OF_CONTENTS;
+
 	return SC_SUCCESS;
 }
