@@ -31,40 +31,163 @@ struct hash_signature_info {
 	CK_MECHANISM_TYPE	hash_mech;
 	CK_MECHANISM_TYPE	sign_mech;
 	sc_pkcs11_mechanism_type_t *hash_type;
-	sc_pkcs11_mechanism_type_t *sign_type;
 };
 
 /* Also used for verification and decryption data */
-struct signature_data {
+struct operation_data {
 	struct sc_pkcs11_object *key;
 	struct hash_signature_info *info;
-	sc_pkcs11_operation_t *	md;
-	CK_BYTE			buffer[4096/8];
+	sc_pkcs11_operation_t *md;
+	CK_BYTE			*buffer;
 	unsigned int	buffer_len;
 };
 
+static struct operation_data *
+new_operation_data()
+{
+	return calloc(1, sizeof(struct operation_data));
+}
+
+static void
+operation_data_release(struct operation_data *data)
+{
+	if (!data)
+		return;
+	sc_pkcs11_release_operation(&data->md);
+	sc_mem_secure_clear_free(data->buffer, data->buffer_len);
+	free(data);
+}
+
+static CK_RV
+signature_data_buffer_append(struct operation_data *data,
+		const CK_BYTE *in, unsigned int in_len)
+{
+	if (!data)
+		return CKR_ARGUMENTS_BAD;
+	if (in_len == 0)
+		return CKR_OK;
+
+	unsigned int new_len = data->buffer_len + in_len;
+	CK_BYTE *new_buffer = sc_mem_secure_alloc(new_len);
+	if (!new_buffer)
+		return CKR_HOST_MEMORY;
+
+	if (data->buffer_len != 0)
+		memcpy(new_buffer, data->buffer, data->buffer_len);
+	memcpy(new_buffer+data->buffer_len, in, in_len);
+
+	sc_mem_secure_clear_free(data->buffer, data->buffer_len);
+	data->buffer = new_buffer;
+	data->buffer_len = new_len;
+	return CKR_OK;
+}
+
+void _update_mech_info(CK_MECHANISM_INFO_PTR mech_info, CK_MECHANISM_INFO_PTR new_mech_info) {
+	if (new_mech_info->ulMaxKeySize > mech_info->ulMaxKeySize)
+		mech_info->ulMaxKeySize = new_mech_info->ulMaxKeySize;
+	if (new_mech_info->ulMinKeySize < mech_info->ulMinKeySize)
+		mech_info->ulMinKeySize = new_mech_info->ulMinKeySize;
+	mech_info->flags |= new_mech_info->flags;
+}
+
+/*
+ * Copy a mechanism
+ */
+static CK_RV
+sc_pkcs11_copy_mechanism(sc_pkcs11_mechanism_type_t *mt,
+				sc_pkcs11_mechanism_type_t **new_mt)
+{
+	int rv;
+
+	*new_mt = calloc(1, sizeof(sc_pkcs11_mechanism_type_t));
+	if (!(*new_mt))
+		return CKR_HOST_MEMORY;
+	
+	memcpy(*new_mt, mt, sizeof(sc_pkcs11_mechanism_type_t));
+	/* mech_data needs specific function for making copy*/
+	if (mt->copy_mech_data
+		&& (rv = mt->copy_mech_data(mt->mech_data, (void **) &(*new_mt)->mech_data)) != CKR_OK) {
+		free(*new_mt);
+		return rv;
+	}
+	
+	return CKR_OK;
+}
+
 /*
  * Register a mechanism
+ * Check whether the mechanism is already in p11card,
+ * if not, make a copy of the given mechanism and store it
+ * in p11card.
  */
 CK_RV
 sc_pkcs11_register_mechanism(struct sc_pkcs11_card *p11card,
-				sc_pkcs11_mechanism_type_t *mt)
+				sc_pkcs11_mechanism_type_t *mt, sc_pkcs11_mechanism_type_t **result_mt)
 {
+	sc_pkcs11_mechanism_type_t *existing_mt;
+	sc_pkcs11_mechanism_type_t *copy_mt = NULL;
 	sc_pkcs11_mechanism_type_t **p;
+	int i, rv;
 
 	if (mt == NULL)
 		return CKR_HOST_MEMORY;
 
+	if ((existing_mt = sc_pkcs11_find_mechanism(p11card, mt->mech, mt->mech_info.flags))) {
+		for (i = 0; i < MAX_KEY_TYPES; i++) {
+			if (existing_mt->key_types[i] == mt->key_types[0]) {
+				/* Mechanism already registered with the same key_type,
+				 * just update it's info */
+				_update_mech_info(&existing_mt->mech_info, &mt->mech_info);
+				return CKR_OK;
+			}
+			if (existing_mt->key_types[i] < 0) {
+				/* There is a free slot for new key type, let's add it and
+				 * update mechanism info */
+				_update_mech_info(&existing_mt->mech_info, &mt->mech_info);
+				/* XXX Should be changed to loop over mt->key_types, if
+				 * we allow to register mechanism with multiple key types
+				 * in one operation */
+				existing_mt->key_types[i] = mt->key_types[0];
+				if (i + 1 < MAX_KEY_TYPES) {
+					existing_mt->key_types[i + 1] = -1;
+				}
+				return CKR_OK;
+			}
+		}
+		sc_log(p11card->card->ctx, "Too many key types in mechanism 0x%lx, more than %d", mt->mech, MAX_KEY_TYPES);
+		return CKR_BUFFER_TOO_SMALL;
+	}
+
 	p = (sc_pkcs11_mechanism_type_t **) realloc(p11card->mechanisms,
 			(p11card->nmechanisms + 2) * sizeof(*p));
 	if (p == NULL)
 		return CKR_HOST_MEMORY;
+	if ((rv = sc_pkcs11_copy_mechanism(mt, &copy_mt)) != CKR_OK) {
+		free(p);
+		return rv;
+	}
 	p11card->mechanisms = p;
-	p[p11card->nmechanisms++] = mt;
+	p[p11card->nmechanisms++] = copy_mt;
 	p[p11card->nmechanisms] = NULL;
+	/* Return registered mechanism for further use */
+	if (result_mt)
+		*result_mt = copy_mt;
 	return CKR_OK;
 }
 
+
+CK_RV
+_validate_key_type(sc_pkcs11_mechanism_type_t *mech, CK_KEY_TYPE key_type) {
+	int i;
+	for (i = 0; i < MAX_KEY_TYPES; i++) {
+		if (mech->key_types[i] < 0)
+			break;
+		if ((CK_KEY_TYPE)mech->key_types[i] == key_type)
+			return CKR_OK;
+	}
+	return CKR_KEY_TYPE_INCONSISTENT;
+}
+
 /*
  * Look up a mechanism
  */
@@ -240,7 +363,7 @@ sc_pkcs11_md_final(struct sc_pkcs11_session *session,
  */
 CK_RV
 sc_pkcs11_sign_init(struct sc_pkcs11_session *session, CK_MECHANISM_PTR pMechanism,
-		    struct sc_pkcs11_object *key, CK_MECHANISM_TYPE key_type)
+		    struct sc_pkcs11_object *key, CK_KEY_TYPE key_type)
 {
 	struct sc_pkcs11_card *p11card;
 	sc_pkcs11_operation_t *operation;
@@ -259,8 +382,9 @@ sc_pkcs11_sign_init(struct sc_pkcs11_session *session, CK_MECHANISM_PTR pMechani
 		LOG_FUNC_RETURN(context, CKR_MECHANISM_INVALID);
 
 	/* See if compatible with key type */
-	if (mt->key_type != key_type)
-		LOG_FUNC_RETURN(context, CKR_KEY_TYPE_INCONSISTENT);
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int) rv);
 
 	if (pMechanism->pParameter &&
 	    pMechanism->ulParameterLen > sizeof(operation->mechanism_params))
@@ -369,12 +493,12 @@ sc_pkcs11_signature_init(sc_pkcs11_operation_t *operation,
 		struct sc_pkcs11_object *key)
 {
 	struct hash_signature_info *info;
-	struct signature_data *data;
+	struct operation_data *data;
 	CK_RV rv;
 	int can_do_it = 0;
 
 	LOG_FUNC_CALLED(context);
-	if (!(data = calloc(1, sizeof(*data))))
+	if (!(data = new_operation_data()))
 		LOG_FUNC_RETURN(context, CKR_HOST_MEMORY);
 	data->info = NULL;
 	data->key = key;
@@ -391,7 +515,7 @@ sc_pkcs11_signature_init(sc_pkcs11_operation_t *operation,
 		}
 		else  {
 			/* Mechanism recognised but cannot be performed by pkcs#15 card, or some general error. */
-			free(data);
+			operation_data_release(data);
 			LOG_FUNC_RETURN(context, (int) rv);
 		}
 	}
@@ -401,7 +525,7 @@ sc_pkcs11_signature_init(sc_pkcs11_operation_t *operation,
 		rv = key->ops->init_params(operation->session, &operation->mechanism);
 		if (rv != CKR_OK) {
 			/* Probably bad arguments */
-			free(data);
+			operation_data_release(data);
 			LOG_FUNC_RETURN(context, (int) rv);
 		}
 	}
@@ -409,7 +533,7 @@ sc_pkcs11_signature_init(sc_pkcs11_operation_t *operation,
 	/* If this is a signature with hash operation,
 	 * and card cannot perform itself signature with hash operation,
 	 * set up the hash operation */
-	info = (struct hash_signature_info *) operation->type->mech_data;
+	info = (struct hash_signature_info *)operation->type->mech_data;
 	if (info != NULL && !can_do_it) {
 		/* Initialize hash operation */
 
@@ -420,7 +544,7 @@ sc_pkcs11_signature_init(sc_pkcs11_operation_t *operation,
 			rv = info->hash_type->md_init(data->md);
 		if (rv != CKR_OK) {
 			sc_pkcs11_release_operation(&data->md);
-			free(data);
+			operation_data_release(data);
 			LOG_FUNC_RETURN(context, (int) rv);
 		}
 		data->info = info;
@@ -434,43 +558,44 @@ static CK_RV
 sc_pkcs11_signature_update(sc_pkcs11_operation_t *operation,
 		CK_BYTE_PTR pPart, CK_ULONG ulPartLen)
 {
-	struct signature_data *data;
+	struct operation_data *data;
+	CK_RV rv;
 
 	LOG_FUNC_CALLED(context);
 	sc_log(context, "data part length %li", ulPartLen);
-	data = (struct signature_data *) operation->priv_data;
+	data = (struct operation_data *)operation->priv_data;
 	if (data->md) {
-		CK_RV rv = data->md->type->md_update(data->md, pPart, ulPartLen);
+		rv = data->md->type->md_update(data->md, pPart, ulPartLen);
 		LOG_FUNC_RETURN(context, (int) rv);
 	}
 
 	/* This signature mechanism operates on the raw data */
-	if (data->buffer_len + ulPartLen > sizeof(data->buffer))
-		LOG_FUNC_RETURN(context, CKR_DATA_LEN_RANGE);
-	memcpy(data->buffer + data->buffer_len, pPart, ulPartLen);
-	data->buffer_len += ulPartLen;
-	LOG_FUNC_RETURN(context, CKR_OK);
+	rv = signature_data_buffer_append(data, pPart, ulPartLen);
+	LOG_FUNC_RETURN(context, (int) rv);
 }
 
 static CK_RV
 sc_pkcs11_signature_final(sc_pkcs11_operation_t *operation,
 		CK_BYTE_PTR pSignature, CK_ULONG_PTR pulSignatureLen)
 {
-	struct signature_data *data;
+	struct operation_data *data;
 	CK_RV rv;
 
 	LOG_FUNC_CALLED(context);
-	data = (struct signature_data *) operation->priv_data;
+	data = (struct operation_data *)operation->priv_data;
 	if (data->md) {
 		sc_pkcs11_operation_t	*md = data->md;
-		CK_ULONG len = sizeof(data->buffer);
+		CK_BYTE hash[64];
+		CK_ULONG len = sizeof(hash);
 
-		rv = md->type->md_final(md, data->buffer, &len);
+		rv = md->type->md_final(md, hash, &len);
 		if (rv == CKR_BUFFER_TOO_SMALL)
 			rv = CKR_FUNCTION_FAILED;
 		if (rv != CKR_OK)
 			LOG_FUNC_RETURN(context, (int) rv);
-		data->buffer_len = (unsigned int) len;
+		rv = signature_data_buffer_append(data, hash, len);
+		if (rv != CKR_OK)
+			LOG_FUNC_RETURN(context, (int) rv);
 	}
 
 	rv = data->key->ops->sign(operation->session, data->key, &operation->mechanism,
@@ -487,7 +612,7 @@ sc_pkcs11_signature_size(sc_pkcs11_operation_t *operation, CK_ULONG_PTR pLength)
 	CK_ATTRIBUTE attr_key_type = { CKA_KEY_TYPE, &key_type, sizeof(key_type) };
 	CK_RV rv;
 
-	key = ((struct signature_data *) operation->priv_data)->key;
+	key = ((struct operation_data *)operation->priv_data)->key;
 	/*
 	 * EC and GOSTR do not have CKA_MODULUS_BITS attribute.
 	 * But other code in framework treats them as if they do.
@@ -525,16 +650,11 @@ sc_pkcs11_signature_size(sc_pkcs11_operation_t *operation, CK_ULONG_PTR pLength)
 }
 
 static void
-sc_pkcs11_signature_release(sc_pkcs11_operation_t *operation)
+sc_pkcs11_operation_release(sc_pkcs11_operation_t *operation)
 {
-	struct signature_data *data;
-
-	data = (struct signature_data *) operation->priv_data;
-	if (!data)
+	if (!operation)
 	    return;
-	sc_pkcs11_release_operation(&data->md);
-	memset(data, 0, sizeof(*data));
-	free(data);
+	operation_data_release(operation->priv_data);
 }
 
 #ifdef ENABLE_OPENSSL
@@ -544,7 +664,7 @@ sc_pkcs11_signature_release(sc_pkcs11_operation_t *operation)
  */
 CK_RV
 sc_pkcs11_verif_init(struct sc_pkcs11_session *session, CK_MECHANISM_PTR pMechanism,
-		struct sc_pkcs11_object *key, CK_MECHANISM_TYPE key_type)
+		struct sc_pkcs11_object *key, CK_KEY_TYPE key_type)
 {
 	struct sc_pkcs11_card *p11card;
 	sc_pkcs11_operation_t *operation;
@@ -561,8 +681,9 @@ sc_pkcs11_verif_init(struct sc_pkcs11_session *session, CK_MECHANISM_PTR pMechan
 		return CKR_MECHANISM_INVALID;
 
 	/* See if compatible with key type */
-	if (mt->key_type != key_type)
-		return CKR_KEY_TYPE_INCONSISTENT;
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int) rv);
 
 	rv = session_start_operation(session, SC_PKCS11_OPERATION_VERIFY, mt, &operation);
 	if (rv != CKR_OK)
@@ -640,10 +761,10 @@ sc_pkcs11_verify_init(sc_pkcs11_operation_t *operation,
 		    struct sc_pkcs11_object *key)
 {
 	struct hash_signature_info *info;
-	struct signature_data *data;
+	struct operation_data *data;
 	CK_RV rv;
 
-	if (!(data = calloc(1, sizeof(*data))))
+	if (!(data = new_operation_data()))
 		return CKR_HOST_MEMORY;
 
 	data->info = NULL;
@@ -698,9 +819,9 @@ static CK_RV
 sc_pkcs11_verify_update(sc_pkcs11_operation_t *operation,
 		    CK_BYTE_PTR pPart, CK_ULONG ulPartLen)
 {
-	struct signature_data *data;
+	struct operation_data *data;
 
-	data = (struct signature_data *) operation->priv_data;
+	data = (struct operation_data *)operation->priv_data;
 	if (data->md) {
 		sc_pkcs11_operation_t	*md = data->md;
 
@@ -708,18 +829,15 @@ sc_pkcs11_verify_update(sc_pkcs11_operation_t *operation,
 	}
 
 	/* This verification mechanism operates on the raw data */
-	if (data->buffer_len + ulPartLen > sizeof(data->buffer))
-		return CKR_DATA_LEN_RANGE;
-	memcpy(data->buffer + data->buffer_len, pPart, ulPartLen);
-	data->buffer_len += ulPartLen;
-	return CKR_OK;
+	CK_RV rv = signature_data_buffer_append(data, pPart, ulPartLen);
+	LOG_FUNC_RETURN(context, (int) rv);
 }
 
 static CK_RV
 sc_pkcs11_verify_final(sc_pkcs11_operation_t *operation,
 			CK_BYTE_PTR pSignature, CK_ULONG ulSignatureLen)
 {
-	struct signature_data *data;
+	struct operation_data *data;
 	struct sc_pkcs11_object *key;
 	unsigned char *pubkey_value = NULL;
 	CK_KEY_TYPE key_type;
@@ -729,7 +847,7 @@ sc_pkcs11_verify_final(sc_pkcs11_operation_t *operation,
 	CK_ATTRIBUTE attr_key_params = {CKA_GOSTR3410_PARAMS, &params, sizeof(params)};
 	CK_RV rv;
 
-	data = (struct signature_data *) operation->priv_data;
+	data = (struct operation_data *)operation->priv_data;
 
 	if (pSignature == NULL)
 		return CKR_ARGUMENTS_BAD;
@@ -773,6 +891,132 @@ done:
 	return rv;
 }
 #endif
+/*
+ * Initialize a encrypting context. When we get here, we know
+ * the key object is capable of encrypt _something_
+ */
+CK_RV
+sc_pkcs11_encr_init(struct sc_pkcs11_session *session,
+		CK_MECHANISM_PTR pMechanism,
+		struct sc_pkcs11_object *key,
+		CK_KEY_TYPE key_type)
+{
+	struct sc_pkcs11_card *p11card;
+	sc_pkcs11_operation_t *operation;
+	sc_pkcs11_mechanism_type_t *mt;
+	CK_RV rv;
+
+	if (!session || !session->slot || !(p11card = session->slot->p11card))
+		return CKR_ARGUMENTS_BAD;
+
+	/* See if we support this mechanism type */
+	mt = sc_pkcs11_find_mechanism(p11card, pMechanism->mechanism, CKF_ENCRYPT);
+	if (mt == NULL)
+		return CKR_MECHANISM_INVALID;
+
+	/* See if compatible with key type */
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int)rv);
+
+	rv = session_start_operation(session, SC_PKCS11_OPERATION_ENCRYPT, mt, &operation);
+	if (rv != CKR_OK)
+		return rv;
+
+	memcpy(&operation->mechanism, pMechanism, sizeof(CK_MECHANISM));
+	if (pMechanism->pParameter) {
+		memcpy(&operation->mechanism_params, pMechanism->pParameter,
+				pMechanism->ulParameterLen);
+		operation->mechanism.pParameter = &operation->mechanism_params;
+	}
+	rv = mt->encrypt_init(operation, key);
+	if (rv != CKR_OK)
+		goto out;
+
+	/* Validate the mechanism parameters */
+	if (key->ops->init_params) {
+		rv = key->ops->init_params(operation->session, &operation->mechanism);
+		if (rv != CKR_OK)
+			goto out;
+	}
+	LOG_FUNC_RETURN(context, (int)rv);
+out:
+	session_stop_operation(session, SC_PKCS11_OPERATION_ENCRYPT);
+	LOG_FUNC_RETURN(context, (int)rv);
+}
+
+CK_RV
+sc_pkcs11_encr(struct sc_pkcs11_session *session,
+		CK_BYTE_PTR pData, CK_ULONG ulDataLen,
+		CK_BYTE_PTR pEncryptedData, CK_ULONG_PTR pulEncryptedDataLen)
+{
+	sc_pkcs11_operation_t *op;
+	CK_RV rv;
+
+	rv = session_get_operation(session, SC_PKCS11_OPERATION_ENCRYPT, &op);
+	if (rv != CKR_OK)
+		return rv;
+
+	rv = op->type->encrypt(op, pData, ulDataLen,
+			pEncryptedData, pulEncryptedDataLen);
+
+	/* application is requesting buffer size ? */
+	if (pEncryptedData == NULL) {
+		/* do not terminate session for CKR_OK */
+		if (rv == CKR_OK)
+			LOG_FUNC_RETURN(context, CKR_OK);
+	} else if (rv == CKR_BUFFER_TOO_SMALL)
+		LOG_FUNC_RETURN(context, CKR_BUFFER_TOO_SMALL);
+
+	session_stop_operation(session, SC_PKCS11_OPERATION_ENCRYPT);
+	LOG_FUNC_RETURN(context, (int)rv);
+}
+
+CK_RV
+sc_pkcs11_encr_update(struct sc_pkcs11_session *session,
+		CK_BYTE_PTR pData, CK_ULONG ulDataLen,
+		CK_BYTE_PTR pEncryptedData, CK_ULONG_PTR pulEncryptedDataLen)
+{
+	sc_pkcs11_operation_t *op;
+	CK_RV rv;
+
+	rv = session_get_operation(session, SC_PKCS11_OPERATION_ENCRYPT, &op);
+	if (rv != CKR_OK)
+		return rv;
+
+	rv = op->type->encrypt_update(op, pData, ulDataLen,
+			pEncryptedData, pulEncryptedDataLen);
+
+	/* terminate session for any error except CKR_BUFFER_TOO_SMALL */
+	if (rv != CKR_OK && rv != CKR_BUFFER_TOO_SMALL)
+		session_stop_operation(session, SC_PKCS11_OPERATION_ENCRYPT);
+	LOG_FUNC_RETURN(context, (int)rv);
+}
+
+CK_RV
+sc_pkcs11_encr_final(struct sc_pkcs11_session *session,
+		CK_BYTE_PTR pEncryptedData, CK_ULONG_PTR pulEncryptedDataLen)
+{
+	sc_pkcs11_operation_t *op;
+	CK_RV rv;
+
+	rv = session_get_operation(session, SC_PKCS11_OPERATION_ENCRYPT, &op);
+	if (rv != CKR_OK)
+		return rv;
+
+	rv = op->type->encrypt_final(op, pEncryptedData, pulEncryptedDataLen);
+
+	/* application is requesting buffer size ? */
+	if (pEncryptedData == NULL) {
+		/* do not terminate session for CKR_OK */
+		if (rv == CKR_OK)
+			LOG_FUNC_RETURN(context, CKR_OK);
+	} else if (rv == CKR_BUFFER_TOO_SMALL)
+		LOG_FUNC_RETURN(context, CKR_BUFFER_TOO_SMALL);
+
+	session_stop_operation(session, SC_PKCS11_OPERATION_ENCRYPT);
+	LOG_FUNC_RETURN(context, (int)rv);
+}
 
 /*
  * Initialize a decryption context. When we get here, we know
@@ -782,7 +1026,7 @@ CK_RV
 sc_pkcs11_decr_init(struct sc_pkcs11_session *session,
 			CK_MECHANISM_PTR pMechanism,
 			struct sc_pkcs11_object *key,
-			CK_MECHANISM_TYPE key_type)
+			CK_KEY_TYPE key_type)
 {
 	struct sc_pkcs11_card *p11card;
 	sc_pkcs11_operation_t *operation;
@@ -799,8 +1043,9 @@ sc_pkcs11_decr_init(struct sc_pkcs11_session *session,
 		return CKR_MECHANISM_INVALID;
 
 	/* See if compatible with key type */
-	if (mt->key_type != key_type)
-		return CKR_KEY_TYPE_INCONSISTENT;
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int) rv);
 
 	rv = session_start_operation(session, SC_PKCS11_OPERATION_DECRYPT, mt, &operation);
 	if (rv != CKR_OK)
@@ -850,6 +1095,52 @@ sc_pkcs11_decr(struct sc_pkcs11_session *session,
 	return rv;
 }
 
+CK_RV
+sc_pkcs11_decr_update(struct sc_pkcs11_session *session,
+		CK_BYTE_PTR pEncryptedData, CK_ULONG ulEncryptedDataLen,
+		CK_BYTE_PTR pData, CK_ULONG_PTR pulDataLen)
+{
+	sc_pkcs11_operation_t *op;
+	CK_RV rv;
+
+	rv = session_get_operation(session, SC_PKCS11_OPERATION_DECRYPT, &op);
+	if (rv != CKR_OK)
+		return rv;
+
+	rv = op->type->decrypt_update(op, pEncryptedData, ulEncryptedDataLen,
+			pData, pulDataLen);
+
+	/* terminate session for any error except CKR_BUFFER_TOO_SMALL */
+	if (rv != CKR_OK && rv != CKR_BUFFER_TOO_SMALL)
+		session_stop_operation(session, SC_PKCS11_OPERATION_DECRYPT);
+	LOG_FUNC_RETURN(context, (int)rv);
+}
+
+CK_RV
+sc_pkcs11_decr_final(struct sc_pkcs11_session *session,
+		CK_BYTE_PTR pData, CK_ULONG_PTR pulDataLen)
+{
+	sc_pkcs11_operation_t *op;
+	CK_RV rv;
+
+	rv = session_get_operation(session, SC_PKCS11_OPERATION_DECRYPT, &op);
+	if (rv != CKR_OK)
+		return rv;
+
+	rv = op->type->decrypt_final(op, pData, pulDataLen);
+
+	/* application is requesting buffer size ? */
+	if (pData == NULL) {
+		/* do not terminate session for CKR_OK */
+		if (rv == CKR_OK)
+			LOG_FUNC_RETURN(context, CKR_OK);
+	} else if (rv == CKR_BUFFER_TOO_SMALL)
+		LOG_FUNC_RETURN(context, CKR_BUFFER_TOO_SMALL);
+
+	session_stop_operation(session, SC_PKCS11_OPERATION_DECRYPT);
+	LOG_FUNC_RETURN(context, (int)rv);
+}
+
 CK_RV
 sc_pkcs11_wrap(struct sc_pkcs11_session *session,
 	CK_MECHANISM_PTR pMechanism,
@@ -873,9 +1164,9 @@ sc_pkcs11_wrap(struct sc_pkcs11_session *session,
 		return CKR_MECHANISM_INVALID;
 
 	/* See if compatible with key type */
-	/* TODO: what if there are several mechanisms with different key types? Should we loop through them? */
-	if (mt->key_type != key_type)
-		return CKR_KEY_TYPE_INCONSISTENT;
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int) rv);
 
 	rv = session_start_operation(session, SC_PKCS11_OPERATION_WRAP, mt, &operation);
 	if (rv != CKR_OK)
@@ -919,9 +1210,9 @@ sc_pkcs11_unwrap(struct sc_pkcs11_session *session,
 		return CKR_MECHANISM_INVALID;
 
 	/* See if compatible with key type */
-	/* TODO: what if there are several mechanisms with different key types? Should we loop through them? */
-	if (mt->key_type != key_type)
-		return CKR_KEY_TYPE_INCONSISTENT;
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int) rv);
 
 	rv = session_start_operation(session, SC_PKCS11_OPERATION_UNWRAP, mt, &operation);
 	if (rv != CKR_OK)
@@ -946,8 +1237,6 @@ sc_pkcs11_unwrap(struct sc_pkcs11_session *session,
 	return rv;
 }
 
-
-
 /* Derive one key from another, and return results in created object */
 CK_RV
 sc_pkcs11_deri(struct sc_pkcs11_session *session,
@@ -981,9 +1270,9 @@ sc_pkcs11_deri(struct sc_pkcs11_session *session,
 		return CKR_MECHANISM_INVALID;
 
 	/* See if compatible with key type */
-	if (mt->key_type != key_type)
-		return CKR_KEY_TYPE_INCONSISTENT;
-
+	rv = _validate_key_type(mt, key_type);
+	if (rv != CKR_OK)
+		LOG_FUNC_RETURN(context, (int) rv);
 
 	rv = session_start_operation(session, SC_PKCS11_OPERATION_DERIVE, mt, &operation);
 	if (rv != CKR_OK)
@@ -1045,6 +1334,140 @@ out:
 	return rv;
 }
 
+/*
+ * Initialize a encrypt operation
+ */
+static CK_RV
+sc_pkcs11_encrypt_init(sc_pkcs11_operation_t *operation,
+		struct sc_pkcs11_object *key)
+{
+	struct operation_data *data;
+	CK_RV rv;
+
+	if (!(data = new_operation_data()))
+		return CKR_HOST_MEMORY;
+
+	data->key = key;
+
+	if (key->ops->can_do) {
+		rv = key->ops->can_do(operation->session, key, operation->type->mech, CKF_ENCRYPT);
+		if ((rv == CKR_OK) || (rv == CKR_FUNCTION_NOT_SUPPORTED)) {
+			/* Mechanism recognized and can be performed by pkcs#15 card or algorithm references not supported */
+		} else {
+			/* Mechanism cannot be performed by pkcs#15 card, or some general error. */
+			free(data);
+			LOG_FUNC_RETURN(context, (int)rv);
+		}
+	}
+
+	operation->priv_data = data;
+
+	/* The last parameter is NULL - this is call to INIT code in underlying functions */
+	return key->ops->encrypt(operation->session,
+			key, &operation->mechanism, NULL, 0, NULL, NULL);
+}
+
+static CK_RV
+sc_pkcs11_encrypt(sc_pkcs11_operation_t *operation,
+		CK_BYTE_PTR pData, CK_ULONG ulDataLen,
+		CK_BYTE_PTR pEncryptedData, CK_ULONG_PTR pulEncryptedDataLen)
+{
+	struct operation_data *data;
+	struct sc_pkcs11_object *key;
+	CK_RV rv;
+	CK_ULONG ulEncryptedDataLen, ulLastEncryptedPartLen;
+
+	/* PKCS#11: If pBuf is not NULL_PTR, then *pulBufLen must contain the size in bytes.. */
+	if (pEncryptedData && !pulEncryptedDataLen)
+		return CKR_ARGUMENTS_BAD;
+
+	ulEncryptedDataLen = pulEncryptedDataLen ? *pulEncryptedDataLen : 0;
+	ulLastEncryptedPartLen = ulEncryptedDataLen;
+
+	data = (struct operation_data *)operation->priv_data;
+
+	key = data->key;
+
+	/* Encrypt (Update) */
+	rv = key->ops->encrypt(operation->session, key, &operation->mechanism,
+			pData, ulDataLen, pEncryptedData, &ulEncryptedDataLen);
+
+	if (pulEncryptedDataLen)
+		*pulEncryptedDataLen = ulEncryptedDataLen;
+
+	if (rv != CKR_OK)
+		return rv;
+
+	/* recalculate buffer space */
+	if (ulEncryptedDataLen <= ulLastEncryptedPartLen)
+		ulLastEncryptedPartLen -= ulEncryptedDataLen;
+	else
+		ulLastEncryptedPartLen = 0;
+
+	/* EncryptFinalize */
+	rv = key->ops->encrypt(operation->session, key, &operation->mechanism,
+			NULL, 0, pEncryptedData + ulEncryptedDataLen, &ulLastEncryptedPartLen);
+
+	if (pulEncryptedDataLen)
+		*pulEncryptedDataLen = ulEncryptedDataLen + ulLastEncryptedPartLen;
+	return rv;
+}
+
+static CK_RV
+sc_pkcs11_encrypt_update(sc_pkcs11_operation_t *operation,
+		CK_BYTE_PTR pPart, CK_ULONG ulPartLen,
+		CK_BYTE_PTR pEncryptedPart, CK_ULONG_PTR pulEncryptedPartLen)
+{
+	struct operation_data *data;
+	struct sc_pkcs11_object *key;
+	CK_RV rv;
+	CK_ULONG ulEncryptedPartLen;
+
+	/* PKCS#11: If pBuf is not NULL_PTR, then *pulBufLen must contain the size in bytes.. */
+	if (pEncryptedPart && !pulEncryptedPartLen)
+		return CKR_ARGUMENTS_BAD;
+
+	ulEncryptedPartLen = pulEncryptedPartLen ? *pulEncryptedPartLen : 0;
+
+	data = (struct operation_data *)operation->priv_data;
+
+	key = data->key;
+
+	rv = key->ops->encrypt(operation->session, key, &operation->mechanism,
+			pPart, ulPartLen, pEncryptedPart, &ulEncryptedPartLen);
+
+	if (pulEncryptedPartLen)
+		*pulEncryptedPartLen = ulEncryptedPartLen;
+	return rv;
+}
+
+static CK_RV
+sc_pkcs11_encrypt_final(sc_pkcs11_operation_t *operation,
+		CK_BYTE_PTR pLastEncryptedPart,
+		CK_ULONG_PTR pulLastEncryptedPartLen)
+{
+	struct operation_data *data;
+	struct sc_pkcs11_object *key;
+	CK_RV rv;
+	CK_ULONG ulLastEncryptedPartLen;
+
+	/* PKCS#11: If pBuf is not NULL_PTR, then *pulBufLen must contain the size in bytes.. */
+	if (pLastEncryptedPart && !pulLastEncryptedPartLen)
+		return CKR_ARGUMENTS_BAD;
+
+	ulLastEncryptedPartLen = pulLastEncryptedPartLen ? *pulLastEncryptedPartLen : 0;
+
+	data = (struct operation_data *)operation->priv_data;
+
+	key = data->key;
+
+	rv = key->ops->encrypt(operation->session, key, &operation->mechanism,
+			NULL, 0, pLastEncryptedPart, &ulLastEncryptedPartLen);
+
+	if (pulLastEncryptedPartLen)
+		*pulLastEncryptedPartLen = ulLastEncryptedPartLen;
+	return rv;
+}
 
 /*
  * Initialize a decrypt operation
@@ -1053,10 +1476,10 @@ static CK_RV
 sc_pkcs11_decrypt_init(sc_pkcs11_operation_t *operation,
 			struct sc_pkcs11_object *key)
 {
-	struct signature_data *data;
+	struct operation_data *data;
 	CK_RV rv;
 
-	if (!(data = calloc(1, sizeof(*data))))
+	if (!(data = new_operation_data()))
 		return CKR_HOST_MEMORY;
 
 	data->key = key;
@@ -1072,9 +1495,11 @@ sc_pkcs11_decrypt_init(sc_pkcs11_operation_t *operation,
 			LOG_FUNC_RETURN(context, (int) rv);
 		}
 	}
-
 	operation->priv_data = data;
-	return CKR_OK;
+
+	/* The last parameter is NULL - this is call to INIT code in underlying functions */
+	return key->ops->decrypt(operation->session,
+			key, &operation->mechanism, NULL, 0, NULL, NULL);
 }
 
 static CK_RV
@@ -1082,16 +1507,104 @@ sc_pkcs11_decrypt(sc_pkcs11_operation_t *operation,
 		CK_BYTE_PTR pEncryptedData, CK_ULONG ulEncryptedDataLen,
 		CK_BYTE_PTR pData, CK_ULONG_PTR pulDataLen)
 {
-	struct signature_data *data;
+	struct operation_data *data;
 	struct sc_pkcs11_object *key;
+	CK_RV rv;
+	CK_ULONG ulDataLen, ulLastDataLen;
 
-	data = (struct signature_data*) operation->priv_data;
+	/* PKCS#11: If pBuf is not NULL_PTR, then *pulBufLen must contain the size in bytes.. */
+	if (pData && !pulDataLen)
+		return CKR_ARGUMENTS_BAD;
+
+	ulDataLen = pulDataLen ? *pulDataLen : 0;
+	ulLastDataLen = ulDataLen;
+
+	data = (struct operation_data *)operation->priv_data;
 
 	key = data->key;
-	return key->ops->decrypt(operation->session,
-				key, &operation->mechanism,
-				pEncryptedData, ulEncryptedDataLen,
-				pData, pulDataLen);
+
+	/* Decrypt */
+	rv = key->ops->decrypt(operation->session, key, &operation->mechanism,
+			pEncryptedData, ulEncryptedDataLen, pData, &ulDataLen);
+
+	if (pulDataLen)
+		*pulDataLen = ulDataLen;
+
+	if (rv != CKR_OK)
+		return rv;
+
+	/* recalculate buffer space */
+	if (ulDataLen <= ulLastDataLen)
+		ulLastDataLen -= ulDataLen;
+	else
+		ulLastDataLen = 0;
+
+	/* DecryptFinalize */
+	rv = key->ops->decrypt(operation->session, key, &operation->mechanism,
+			NULL, 0, pData + ulDataLen, &ulLastDataLen);
+	if (pulDataLen)
+		*pulDataLen = ulDataLen + ulLastDataLen;
+	return rv;
+}
+
+static CK_RV
+sc_pkcs11_decrypt_update(sc_pkcs11_operation_t *operation,
+		CK_BYTE_PTR pEncryptedPart, CK_ULONG ulEncryptedPartLen,
+		CK_BYTE_PTR pPart, CK_ULONG_PTR pulPartLen)
+{
+	struct operation_data *data;
+	struct sc_pkcs11_object *key;
+	CK_RV rv;
+	CK_ULONG ulPartLen;
+
+	/* PKCS#11: If pBuf is not NULL_PTR, then *pulBufLen must contain the size in bytes.. */
+	if (pPart && !pulPartLen)
+		return CKR_ARGUMENTS_BAD;
+
+	ulPartLen = pulPartLen ? *pulPartLen : 0;
+
+	data = (struct operation_data *)operation->priv_data;
+
+	key = data->key;
+
+	rv = key->ops->decrypt(operation->session,
+			key, &operation->mechanism,
+			pEncryptedPart, ulEncryptedPartLen,
+			pPart, &ulPartLen);
+
+	if (pulPartLen)
+		*pulPartLen = ulPartLen;
+	return rv;
+}
+
+static CK_RV
+sc_pkcs11_decrypt_final(sc_pkcs11_operation_t *operation,
+		CK_BYTE_PTR pLastPart,
+		CK_ULONG_PTR pulLastPartLen)
+{
+	struct operation_data *data;
+	struct sc_pkcs11_object *key;
+	CK_RV rv;
+	CK_ULONG ulLastPartLen;
+
+	/* PKCS#11: If pBuf is not NULL_PTR, then *pulBufLen must contain the size in bytes.. */
+	if (pLastPart && !pulLastPartLen)
+		return CKR_ARGUMENTS_BAD;
+
+	ulLastPartLen = pulLastPartLen ? *pulLastPartLen : 0;
+
+	data = (struct operation_data *)operation->priv_data;
+
+	key = data->key;
+
+	rv = key->ops->decrypt(operation->session,
+			key, &operation->mechanism,
+			NULL, 0,
+			pLastPart, &ulLastPartLen);
+
+	if (pulLastPartLen)
+		*pulLastPartLen = ulLastPartLen;
+	return rv;
 }
 
 static CK_RV
@@ -1150,7 +1663,8 @@ sc_pkcs11_new_fw_mechanism(CK_MECHANISM_TYPE mech,
 				CK_MECHANISM_INFO_PTR pInfo,
 				CK_KEY_TYPE key_type,
 				const void *priv_data,
-				void (*free_priv_data)(const void *priv_data))
+				void (*free_priv_data)(const void *priv_data),
+				CK_RV (*copy_priv_data)(const void *mech_data, void **new_data))
 {
 	sc_pkcs11_mechanism_type_t *mt;
 
@@ -1159,12 +1673,14 @@ sc_pkcs11_new_fw_mechanism(CK_MECHANISM_TYPE mech,
 		return mt;
 	mt->mech = mech;
 	mt->mech_info = *pInfo;
-	mt->key_type = key_type;
+	mt->key_types[0] = (int)key_type;
+	mt->key_types[1] = -1;
 	mt->mech_data = priv_data;
 	mt->free_mech_data = free_priv_data;
+	mt->copy_mech_data = copy_priv_data;
 	mt->obj_size = sizeof(sc_pkcs11_operation_t);
 
-	mt->release = sc_pkcs11_signature_release;
+	mt->release = sc_pkcs11_operation_release;
 
 	if (pInfo->flags & CKF_SIGN) {
 		mt->sign_init = sc_pkcs11_signature_init;
@@ -1189,11 +1705,29 @@ sc_pkcs11_new_fw_mechanism(CK_MECHANISM_TYPE mech,
 	if (pInfo->flags & CKF_DECRYPT) {
 		mt->decrypt_init = sc_pkcs11_decrypt_init;
 		mt->decrypt = sc_pkcs11_decrypt;
+		mt->decrypt_update = sc_pkcs11_decrypt_update;
+		mt->decrypt_final = sc_pkcs11_decrypt_final;
+	}
+	if (pInfo->flags & CKF_ENCRYPT) {
+		mt->encrypt_init = sc_pkcs11_encrypt_init;
+		mt->encrypt = sc_pkcs11_encrypt;
+		mt->encrypt_update = sc_pkcs11_encrypt_update;
+		mt->encrypt_final = sc_pkcs11_encrypt_final;
 	}
 
 	return mt;
 }
 
+void sc_pkcs11_free_mechanism(sc_pkcs11_mechanism_type_t **mt)
+{
+	if (!mt || !(*mt))
+		return;
+	if ((*mt)->free_mech_data)
+		(*mt)->free_mech_data((*mt)->mech_data);
+	free(*mt);
+	*mt = NULL;
+}
+
 /*
  * Register generic mechanisms
  */
@@ -1211,6 +1745,19 @@ void free_info(const void *info)
 	free((void *) info);
 }
 
+CK_RV copy_hash_signature_info(const void *mech_data, void **new_data)
+{
+	if (mech_data == NULL || new_data == NULL)
+		return CKR_ARGUMENTS_BAD;
+
+	*new_data = calloc(1, sizeof(struct hash_signature_info));
+	if (!(*new_data))
+		return CKR_HOST_MEMORY;
+	
+	memcpy(*new_data, mech_data, sizeof(struct hash_signature_info));
+	return CKR_OK;
+}
+
 /*
  * Register a sign+hash algorithm derived from an algorithm supported
  * by the token + a software hash mechanism
@@ -1223,9 +1770,13 @@ sc_pkcs11_register_sign_and_hash_mechanism(struct sc_pkcs11_card *p11card,
 {
 	sc_pkcs11_mechanism_type_t *hash_type, *new_type;
 	struct hash_signature_info *info;
-	CK_MECHANISM_INFO mech_info = sign_type->mech_info;
+	CK_MECHANISM_INFO mech_info;
 	CK_RV rv;
 
+	if (!sign_type)
+		return CKR_MECHANISM_INVALID;
+	mech_info = sign_type->mech_info;
+
 	if (!(hash_type = sc_pkcs11_find_mechanism(p11card, hash_mech, CKF_DIGEST)))
 		return CKR_MECHANISM_INVALID;
 
@@ -1237,22 +1788,18 @@ sc_pkcs11_register_sign_and_hash_mechanism(struct sc_pkcs11_card *p11card,
 		return CKR_HOST_MEMORY;
 
 	info->mech = mech;
-	info->sign_type = sign_type;
 	info->hash_type = hash_type;
 	info->sign_mech = sign_type->mech;
 	info->hash_mech = hash_mech;
 
-	new_type = sc_pkcs11_new_fw_mechanism(mech, &mech_info, sign_type->key_type, info, free_info);
+	new_type = sc_pkcs11_new_fw_mechanism(mech, &mech_info, sign_type->key_types[0], info, free_info, copy_hash_signature_info);
 	if (!new_type) {
 		free_info(info);
 		return CKR_HOST_MEMORY;
 	}
 
-	rv = sc_pkcs11_register_mechanism(p11card, new_type);
-	if (CKR_OK != rv) {
-		new_type->free_mech_data(new_type->mech_data);
-		free(new_type);
-	}
+	rv = sc_pkcs11_register_mechanism(p11card, new_type, NULL);
+	sc_pkcs11_free_mechanism(&new_type);
 
 	return rv;
 }
