@@ -37,21 +37,6 @@ static struct sc_to_cryptoki_error_conversion sc_to_cryptoki_error_map[]  = {
 	{ "C_GenerateKeyPair",	SC_ERROR_INVALID_PIN_LENGTH,	CKR_GENERAL_ERROR },
 	{ "C_Sign",		SC_ERROR_NOT_ALLOWED,		CKR_FUNCTION_FAILED},
 	{ "C_Decrypt",		SC_ERROR_NOT_ALLOWED,		CKR_FUNCTION_FAILED},
-	{ "C_GetTokenInfo",		SC_ERROR_INVALID_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetTokenInfo",		SC_ERROR_WRONG_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetTokenInfo",		SC_ERROR_NO_CARD_SUPPORT,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetMechanismList",		SC_ERROR_INVALID_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetMechanismList",		SC_ERROR_WRONG_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetMechanismList",		SC_ERROR_NO_CARD_SUPPORT,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetMechanismInfo",		SC_ERROR_INVALID_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetMechanismInfo",		SC_ERROR_WRONG_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_GetMechanismInfo",		SC_ERROR_NO_CARD_SUPPORT,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_InitToken",		SC_ERROR_INVALID_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_InitToken",		SC_ERROR_WRONG_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_InitToken",		SC_ERROR_NO_CARD_SUPPORT,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_OpenSession",		SC_ERROR_INVALID_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_OpenSession",		SC_ERROR_WRONG_CARD,		CKR_TOKEN_NOT_RECOGNIZED},
-	{ "C_OpenSession",		SC_ERROR_NO_CARD_SUPPORT,		CKR_TOKEN_NOT_RECOGNIZED},
 	{NULL, 0, 0}
 };
 
@@ -102,8 +87,7 @@ static CK_RV sc_to_cryptoki_error_common(int rc)
 	case SC_ERROR_INVALID_CARD:
 	case SC_ERROR_WRONG_CARD:
 	case SC_ERROR_NO_CARD_SUPPORT:
-		/* CKR_TOKEN_NOT_RECOGNIZED is handled in sc_to_cryptoki_error() */
-		return CKR_FUNCTION_NOT_SUPPORTED;
+		return CKR_TOKEN_NOT_RECOGNIZED;
 	case SC_ERROR_WRONG_LENGTH:
 		return CKR_DATA_LEN_RANGE;
 	case SC_ERROR_INVALID_PIN_LENGTH:
@@ -131,6 +115,8 @@ static CK_RV sc_to_cryptoki_error_common(int rc)
 		return CKR_DEVICE_MEMORY;
 	case SC_ERROR_MEMORY_FAILURE:	/* EEPROM has failed */
 		return CKR_DEVICE_ERROR;
+	case SC_ERROR_WRONG_PADDING:
+		return CKR_ENCRYPTED_DATA_INVALID;
 	}
 	return CKR_GENERAL_ERROR;
 }
@@ -232,8 +218,7 @@ CK_RV push_login_state(struct sc_pkcs11_slot *slot,
 err:
 	if (login) {
 		if (login->pPin) {
-			sc_mem_clear(login->pPin, login->ulPinLen);
-			sc_mem_secure_free(login->pPin, login->ulPinLen);
+			sc_mem_secure_clear_free(login->pPin, login->ulPinLen);
 		}
 		free(login);
 	}
@@ -248,8 +233,7 @@ void pop_login_state(struct sc_pkcs11_slot *slot)
 		if (size > 0) {
 			struct sc_pkcs11_login *login = list_get_at(&slot->logins, size-1);
 			if (login) {
-				sc_mem_clear(login->pPin, login->ulPinLen);
-				sc_mem_secure_free(login->pPin, login->ulPinLen);
+				sc_mem_secure_clear_free(login->pPin, login->ulPinLen);
 				free(login);
 			}
 			if (0 > list_delete_at(&slot->logins, size-1))
@@ -263,8 +247,7 @@ void pop_all_login_states(struct sc_pkcs11_slot *slot)
 	if (sc_pkcs11_conf.atomic && slot) {
 		struct sc_pkcs11_login *login = list_fetch(&slot->logins);
 		while (login) {
-			sc_mem_clear(login->pPin, login->ulPinLen);
-			sc_mem_secure_free(login->pPin, login->ulPinLen);
+			sc_mem_secure_clear_free(login->pPin, login->ulPinLen);
 			free(login);
 			login = list_fetch(&slot->logins);
 		}
