@@ -37,7 +37,10 @@
 #include "sc-pkcs11.h"
 #include "ui/notify.h"
 
+#ifdef ENABLE_OPENSSL
+#include <openssl/crypto.h>
 #include "libopensc/sc-ossl-compat.h"
+#endif
 #ifdef ENABLE_OPENPACE
 #include <eac/eac.h>
 #endif
@@ -55,11 +58,17 @@ pid_t initialized_pid = (pid_t)-1;
 #endif
 static int in_finalize = 0;
 extern CK_FUNCTION_LIST pkcs11_function_list;
+extern CK_FUNCTION_LIST_3_0 pkcs11_function_list_3_0;
 
 #ifdef PKCS11_THREAD_LOCKING
 
 #if defined(HAVE_PTHREAD)
 
+/* mutex used to control C_Initilize creation of mutexes */
+static pthread_mutex_t c_initialize_m = PTHREAD_MUTEX_INITIALIZER;
+#define C_INITIALIZE_M_LOCK  pthread_mutex_lock(&c_initialize_m);
+#define C_INITIALIZE_M_UNLOCK pthread_mutex_unlock(&c_initialize_m);
+
 CK_RV mutex_create(void **mutex)
 {
 	pthread_mutex_t *m;
@@ -100,6 +109,9 @@ static CK_C_INITIALIZE_ARGS _def_locks = {
 #define HAVE_OS_LOCKING
 
 #elif defined(_WIN32)
+CRITICAL_SECTION c_initialize_cs = {0};
+#define C_INITIALIZE_M_LOCK EnterCriticalSection(&c_initialize_cs);
+#define C_INITIALIZE_M_UNLOCK LeaveCriticalSection(&c_initialize_cs);
 
 CK_RV mutex_create(void **mutex)
 {
@@ -140,9 +152,15 @@ static CK_C_INITIALIZE_ARGS _def_locks = {
 
 #endif
 
+#else /* PKCS11_THREAD_LOCKING */
+#define C_INITIALIZE_M_LOCK
+#define C_INITIALIZE_M_UNLOCK
+
 #endif /* PKCS11_THREAD_LOCKING */
 
 static CK_C_INITIALIZE_ARGS_PTR	global_locking;
+static CK_C_INITIALIZE_ARGS app_locking = {
+	NULL, NULL, NULL, NULL, 0, NULL };
 static void *global_lock = NULL;
 #ifdef HAVE_OS_LOCKING
 static CK_C_INITIALIZE_ARGS_PTR default_mutex_funcs = &_def_locks;
@@ -220,6 +238,9 @@ __attribute__((constructor))
 #endif
 int module_init()
 {
+#ifdef _WIN32
+	InitializeCriticalSection(&c_initialize_cs);
+#endif
 	sc_notify_init();
 	return 1;
 }
@@ -230,11 +251,14 @@ __attribute__((destructor))
 int module_close()
 {
 	sc_notify_close();
-#if defined(ENABLE_OPENSSL) && defined(OPENSSL_SECURE_MALLOC_SIZE)
+#if defined(ENABLE_OPENSSL) && defined(OPENSSL_SECURE_MALLOC_SIZE) && !defined(LIBRESSL_VERSION_NUMBER)
 	CRYPTO_secure_malloc_done();
 #endif
 #ifdef ENABLE_OPENPACE
 	EAC_cleanup();
+#endif
+#ifdef _WIN32
+	DeleteCriticalSection(&c_initialize_cs);
 #endif
 	return 1;
 }
@@ -282,8 +306,12 @@ CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 	in_finalize = 0;
 #endif
 
+	/* protect from multiple threads tryng to setup locking */
+	C_INITIALIZE_M_LOCK
+
 	if (context != NULL) {
 		sc_log(context, "C_Initialize(): Cryptoki already initialized\n");
+		C_INITIALIZE_M_UNLOCK
 		return CKR_CRYPTOKI_ALREADY_INITIALIZED;
 	}
 
@@ -324,7 +352,7 @@ CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 
 out:
 	if (context != NULL)
-		sc_log(context, "C_Initialize() = %s", lookup_enum ( RV_T, rv ));
+		SC_LOG_RV("C_Initialize() = %s", rv);
 
 	if (rv != CKR_OK) {
 		if (context != NULL) {
@@ -335,6 +363,9 @@ out:
 		sc_pkcs11_free_lock();
 	}
 
+	/* protect from multiple threads tryng to setup locking */
+	C_INITIALIZE_M_UNLOCK
+
 	return rv;
 }
 
@@ -388,7 +419,7 @@ CK_RV C_Finalize(CK_VOID_PTR pReserved)
 	return rv;
 }
 
-CK_RV C_GetInfo(CK_INFO_PTR pInfo)
+CK_RV get_info_version(CK_INFO_PTR pInfo, CK_VERSION version)
 {
 	CK_RV rv = CKR_OK;
 
@@ -402,8 +433,8 @@ CK_RV C_GetInfo(CK_INFO_PTR pInfo)
 	sc_log(context, "C_GetInfo()");
 
 	memset(pInfo, 0, sizeof(CK_INFO));
-	pInfo->cryptokiVersion.major = 2;
-	pInfo->cryptokiVersion.minor = 20;
+	pInfo->cryptokiVersion.major = version.major;
+	pInfo->cryptokiVersion.minor = version.minor;
 	strcpy_bp(pInfo->manufacturerID,
 		  OPENSC_VS_FF_COMPANY_NAME,
 		  sizeof(pInfo->manufacturerID));
@@ -417,6 +448,20 @@ CK_RV C_GetInfo(CK_INFO_PTR pInfo)
 	return rv;
 }
 
+CK_RV C_GetInfoV2(CK_INFO_PTR pInfo)
+{
+	CK_VERSION v = {2, 20};
+
+	return get_info_version(pInfo, v);
+}
+
+CK_RV C_GetInfo(CK_INFO_PTR pInfo)
+{
+	CK_VERSION v = {3, 0};
+
+	return get_info_version(pInfo, v);
+}
+
 CK_RV C_GetFunctionList(CK_FUNCTION_LIST_PTR_PTR ppFunctionList)
 {
 	if (ppFunctionList == NULL_PTR)
@@ -479,9 +524,10 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 		 * - any slot with token;
 		 * - any slot that has already been seen;
 		 */
-		if ((!tokenPresent && slot->reader != prev_reader)
-				|| (slot->slot_info.flags & CKF_TOKEN_PRESENT)
-				|| (slot->flags & SC_PKCS11_SLOT_FLAG_SEEN)) {
+		if ((!tokenPresent &&
+				(slot->reader != prev_reader ||
+			 	 slot->flags & SC_PKCS11_SLOT_FLAG_SEEN))
+				|| slot->slot_info.flags & CKF_TOKEN_PRESENT) {
 			found[numMatches++] = slot->id;
 			slot->flags |= SC_PKCS11_SLOT_FLAG_SEEN;
 		}
@@ -548,6 +594,7 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 {
 	struct sc_pkcs11_slot *slot = NULL;
 	sc_timestamp_t now;
+	const char *name;
 	CK_RV rv;
 
 	if (pInfo == NULL_PTR)
@@ -570,7 +617,7 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 
 	rv = slot_get_slot(slotID, &slot);
 	DEBUG_VSS(slot, "C_GetSlotInfo found");
-	sc_log(context, "C_GetSlotInfo() get slot rv %s", lookup_enum( RV_T, rv));
+	SC_LOG_RV("C_GetSlotInfo() get slot rv %s", rv);
 	if (rv == CKR_OK) {
 		if (slot->reader == NULL) {
 			rv = CKR_TOKEN_NOT_PRESENT;
@@ -597,7 +644,12 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 		memcpy(pInfo, &slot->slot_info, sizeof(CK_SLOT_INFO));
 
 	sc_log(context, "C_GetSlotInfo() flags 0x%lX", pInfo->flags);
-	sc_log(context, "C_GetSlotInfo(0x%lx) = %s", slotID, lookup_enum( RV_T, rv));
+	
+	name = lookup_enum(RV_T, rv);
+	if (name)
+		sc_log(context, "C_GetSlotInfo(0x%lx) = %s", slotID, name);
+	else
+		sc_log(context, "C_GetSlotInfo(0x%lx) = 0x%08lX", slotID, rv);
 	sc_pkcs11_unlock();
 	return rv;
 }
@@ -759,11 +811,99 @@ out:
 		sc_wait_for_event(context, 0, NULL, NULL, -1, &reader_states);
 	}
 
-	sc_log(context, "C_WaitForSlotEvent() = %s", lookup_enum (RV_T, rv));
+	SC_LOG_RV("C_WaitForSlotEvent() = %s", rv);
 	sc_pkcs11_unlock();
 	return rv;
 }
 
+/*
+ * Interfaces
+ */
+#define NUM_INTERFACES 2
+#define DEFAULT_INTERFACE 0
+CK_INTERFACE interfaces[NUM_INTERFACES] = {
+	{"PKCS 11", (void *)&pkcs11_function_list_3_0, 0},
+	{"PKCS 11", (void *)&pkcs11_function_list, 0}
+};
+
+CK_RV C_GetInterfaceList(CK_INTERFACE_PTR pInterfacesList,  /* returned interfaces */
+			 CK_ULONG_PTR pulCount)         /* number of interfaces returned */
+{
+	sc_log(context, "C_GetInterfaceList()");
+
+	if (pulCount == NULL_PTR)
+		return CKR_ARGUMENTS_BAD;
+
+	if (pInterfacesList == NULL_PTR) {
+		*pulCount = NUM_INTERFACES;
+		sc_log(context, "was only a size inquiry (%lu)\n", *pulCount);
+		return CKR_OK;
+	}
+
+	if (*pulCount < NUM_INTERFACES) {
+		sc_log(context, "buffer was too small (needed %d)\n", NUM_INTERFACES);
+		*pulCount = NUM_INTERFACES;
+		return CKR_BUFFER_TOO_SMALL;
+	}
+
+	memcpy(pInterfacesList, interfaces, NUM_INTERFACES * sizeof(CK_INTERFACE));
+	*pulCount = NUM_INTERFACES;
+
+	sc_log(context, "returned %lu interfaces\n", *pulCount);
+	return CKR_OK;
+}
+
+CK_RV C_GetInterface(CK_UTF8CHAR_PTR pInterfaceName, /* name of the interface */
+                     CK_VERSION_PTR pVersion,       /* version of the interface */
+                     CK_INTERFACE_PTR_PTR ppInterface,    /* returned interface */
+                     CK_FLAGS flags)           /* flags controlling the semantics
+						* of the interface */
+{
+	int i;
+
+	sc_log(context, "C_GetInterface(%s)",
+		pInterfaceName == NULL_PTR ? "<default>" : (char *)pInterfaceName);
+
+	if (ppInterface == NULL) {
+		return CKR_ARGUMENTS_BAD;
+	}
+
+	if (pInterfaceName == NULL_PTR) {
+		/* return default interface */
+		*ppInterface = &interfaces[DEFAULT_INTERFACE];
+		sc_log(context, "Returning default interface\n");
+		return CKR_OK;
+	}
+
+	for (i = 0; i < NUM_INTERFACES; i++) {
+		CK_VERSION_PTR interface_version = (CK_VERSION_PTR)interfaces[i].pFunctionList;
+
+		/* The interface name is not null here */
+		if (strcmp((char *)pInterfaceName, interfaces[i].pInterfaceName) != 0) {
+			continue;
+		}
+		/* If version is not null, it must match */
+		if (pVersion != NULL_PTR && (pVersion->major != interface_version->major ||
+		    pVersion->minor != interface_version->minor)) {
+			continue;
+		}
+		/* If any flags specified, it must be supported by the interface */
+		if ((flags & interfaces[i].flags) != flags) {
+			continue;
+		}
+		*ppInterface = &interfaces[i];
+		sc_log(context, "Returning interface %s\n", (*ppInterface)->pInterfaceName);
+		return CKR_OK;
+	}
+	sc_log(context, "Interface not found: %s, version=%d.%d, flags=%lu\n",
+		pInterfaceName ? (char *)pInterfaceName : "<null>",
+		pVersion ? (*pVersion).major : 0,
+		pVersion ? (*pVersion).minor : 0,
+		flags);
+
+	return CKR_ARGUMENTS_BAD;
+}
+
 /*
  * Locking functions
  */
@@ -785,6 +925,8 @@ sc_pkcs11_init_lock(CK_C_INITIALIZE_ARGS_PTR args)
 	if (args->pReserved != NULL_PTR)
 		return CKR_ARGUMENTS_BAD;
 
+	app_locking = *args;
+
 	/* If the app tells us OS locking is okay,
 	 * use that. Otherwise use the supplied functions.
 	 */
@@ -800,13 +942,13 @@ sc_pkcs11_init_lock(CK_C_INITIALIZE_ARGS_PTR args)
 	/* Based on PKCS#11 v2.11 11.4 */
 	if (applock && oslock) {
 		/* Shall be used in threaded environment, prefer app provided locking */
-		global_locking = args;
+		global_locking = &app_locking;
 	} else if (!applock && oslock) {
 		/* Shall be used in threaded environment, must use operating system locking */
 		global_locking = default_mutex_funcs;
 	} else if (applock && !oslock) {
 		/* Shall be used in threaded environment, must use app provided locking */
-		global_locking = args;
+		global_locking = &app_locking;
 	} else if (!applock && !oslock) {
 		/* Shall not be used in threaded environment, use operating system locking */
 		global_locking = default_mutex_funcs;
@@ -877,7 +1019,80 @@ void sc_pkcs11_free_lock(void)
 }
 
 CK_FUNCTION_LIST pkcs11_function_list = {
-	{ 2, 11 }, /* Note: NSS/Firefox ignores this version number and uses C_GetInfo() */
+	{ 2, 20 }, /* Note: NSS/Firefox ignores this version number and uses C_GetInfo() */
+	C_Initialize,
+	C_Finalize,
+	C_GetInfoV2,
+	C_GetFunctionList,
+	C_GetSlotList,
+	C_GetSlotInfo,
+	C_GetTokenInfo,
+	C_GetMechanismList,
+	C_GetMechanismInfo,
+	C_InitToken,
+	C_InitPIN,
+	C_SetPIN,
+	C_OpenSession,
+	C_CloseSession,
+	C_CloseAllSessions,
+	C_GetSessionInfo,
+	C_GetOperationState,
+	C_SetOperationState,
+	C_Login,
+	C_Logout,
+	C_CreateObject,
+	C_CopyObject,
+	C_DestroyObject,
+	C_GetObjectSize,
+	C_GetAttributeValue,
+	C_SetAttributeValue,
+	C_FindObjectsInit,
+	C_FindObjects,
+	C_FindObjectsFinal,
+	C_EncryptInit,
+	C_Encrypt,
+	C_EncryptUpdate,
+	C_EncryptFinal,
+	C_DecryptInit,
+	C_Decrypt,
+	C_DecryptUpdate,
+	C_DecryptFinal,
+	C_DigestInit,
+	C_Digest,
+	C_DigestUpdate,
+	C_DigestKey,
+	C_DigestFinal,
+	C_SignInit,
+	C_Sign,
+	C_SignUpdate,
+	C_SignFinal,
+	C_SignRecoverInit,
+	C_SignRecover,
+	C_VerifyInit,
+	C_Verify,
+	C_VerifyUpdate,
+	C_VerifyFinal,
+	C_VerifyRecoverInit,
+	C_VerifyRecover,
+	C_DigestEncryptUpdate,
+	C_DecryptDigestUpdate,
+	C_SignEncryptUpdate,
+	C_DecryptVerifyUpdate,
+	C_GenerateKey,
+	C_GenerateKeyPair,
+	C_WrapKey,
+	C_UnwrapKey,
+	C_DeriveKey,
+	C_SeedRandom,
+	C_GenerateRandom,
+	C_GetFunctionStatus,
+	C_CancelFunction,
+	C_WaitForSlotEvent
+};
+
+/* Returned from getInterface */
+CK_FUNCTION_LIST_3_0 pkcs11_function_list_3_0 = {
+	{ 3, 0 },
 	C_Initialize,
 	C_Finalize,
 	C_GetInfo,
@@ -945,5 +1160,29 @@ CK_FUNCTION_LIST pkcs11_function_list = {
 	C_GenerateRandom,
 	C_GetFunctionStatus,
 	C_CancelFunction,
-	C_WaitForSlotEvent
+	C_WaitForSlotEvent,
+	C_GetInterfaceList,
+	C_GetInterface,
+	C_LoginUser,
+	C_SessionCancel,
+	C_MessageEncryptInit,
+	C_EncryptMessage,
+	C_EncryptMessageBegin,
+	C_EncryptMessageNext,
+	C_MessageEncryptFinal,
+	C_MessageDecryptInit,
+	C_DecryptMessage,
+	C_DecryptMessageBegin,
+	C_DecryptMessageNext,
+	C_MessageDecryptFinal,
+	C_MessageSignInit,
+	C_SignMessage,
+	C_SignMessageBegin,
+	C_SignMessageNext,
+	C_MessageSignFinal,
+	C_MessageVerifyInit,
+	C_VerifyMessage,
+	C_VerifyMessageBegin,
+	C_VerifyMessageNext,
+	C_MessageVerifyFinal
 };
