@@ -130,6 +130,8 @@ static int entersafe_init_card(sc_profile_t *profile, sc_pkcs15_card_t *p15card)
 
 
 		 /* fill file by 0 */
+		 if (size > MAX_FILE_SIZE)
+			 LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_DATA, "Initialize EF(DIR) failed with file size too large");
 		 buff = calloc(1,size);
 		 if(!buff)
 			  SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_SUCCESS);
@@ -168,7 +170,7 @@ static int entersafe_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card
 		 memcpy(df_data.data.df.aid,df->name,df->namelen);
 
 		 ret = sc_card_ctl(card, SC_CARDCTL_ENTERSAFE_CREATE_FILE, &df_data);
-		 LOG_TEST_RET(card->ctx,ret,"Crate DF failed");
+		 LOG_TEST_RET(card->ctx,ret,"Create DF failed");
 	}
 
 	{/* GPKF */
@@ -356,7 +358,7 @@ static int entersafe_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 
 	if ( key->algorithm != SC_ALGORITHM_RSA )
 	{
-		 /* ignore DSA keys */
+		 /* ignore non-RSA keys */
 		 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE,SC_ERROR_INVALID_ARGUMENTS);
 	}
 
