@@ -168,6 +168,8 @@ myeid_erase_card(struct sc_profile *profile, struct sc_pkcs15_card *p15card) {
 	/* ACLs are not actives if file is not in the operational state */
 	if (mf->status == SC_FILE_STATUS_ACTIVATED)
 		r = sc_pkcs15init_authenticate(profile, p15card, mf, SC_AC_OP_DELETE);
+	if (r < 0)
+		sc_file_free(mf);
 	LOG_TEST_RET(ctx, r, "'DELETE' authentication failed on MF");
 
 	data_obj.P1 = 0x01;
@@ -176,6 +178,7 @@ myeid_erase_card(struct sc_profile *profile, struct sc_pkcs15_card *p15card) {
 	data_obj.DataLen = sizeof (data);
 
 	r = sc_card_ctl(p15card->card, SC_CARDCTL_MYEID_PUTDATA, &data_obj);
+	sc_file_free(mf);
 
 	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
@@ -414,6 +417,11 @@ myeid_new_file(sc_profile_t *profile, sc_card_t *card,
 	file->id += num;
 	p = &file->path;
 	*p = profile->df_info->file->path;
+	if (p->len >= SC_MAX_PATH_SIZE - 2) {
+		sc_log(card->ctx, "Wrong path length");
+		sc_file_free(file);
+		return SC_ERROR_INTERNAL;
+	}
 	p->value[p->len++] = (u8) (file->id / 256);
 	p->value[p->len++] = (u8) (file->id % 256);
 
@@ -595,8 +603,10 @@ myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	r = myeid_new_file(profile, card, object->type, *key_reference, &file);
 	LOG_TEST_RET(ctx, r, "Cannot get new MyEID key file");
 
-	if (!file || !file->path.len)
+	if (!file || !file->path.len || file->path.len > SC_MAX_PATH_SIZE) {
+		sc_file_free(file);
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Cannot determine key file");
+	}
 
 	sc_log(ctx, "Key file size %d", keybits);
 	file->size = keybits;
@@ -810,6 +820,8 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	LOG_TEST_RET(ctx, r, "Cannot generate key: failed to select key file");
 
 	r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_GENERATE);
+	if (r < 0)
+		sc_file_free(file);
 	LOG_TEST_RET(ctx, r, "No authorisation to generate private key");
 
 	/* Fill in data structure */
@@ -826,6 +838,8 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 
 	/* Generate the key  */
 	r = sc_card_ctl(card, SC_CARDCTL_MYEID_GENERATE_STORE_KEY, &args);
+	if (r < 0)
+		sc_file_free(file);
 	LOG_TEST_RET(ctx, r, "Card control 'MYEID_GENERATE_STORE_KEY' failed");
 
 	/* Key pair generation -> collect public key info */
@@ -842,6 +856,8 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 
 			/* Get public key modulus */
 			r = sc_select_file(card, &file->path, NULL);
+			sc_file_free(file);
+			file = NULL;
 			LOG_TEST_RET(ctx, r, "Cannot get key modulus: select key file failed");
 
 			data_obj.P1 = 0x01;
@@ -867,6 +883,8 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 			pubkey->algorithm = SC_ALGORITHM_EC;
 
 			r = sc_select_file(card, &file->path, NULL);
+			sc_file_free(file);
+			file = NULL;
 			LOG_TEST_RET(ctx, r, "Cannot get public key: select key file failed");
 
 			data_obj.P1 = 0x01;
@@ -906,9 +924,14 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 			pubkey->u.ec.params.der.len = 0;
 
 			pubkey->u.ec.params.named_curve = strdup(ecparams->named_curve);
-			if (!pubkey->u.ec.params.named_curve)
+			if (!pubkey->u.ec.params.named_curve) {
+				free(pubkey->u.ec.ecpointQ.value);
 				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+			}
+
 			r = sc_pkcs15_fix_ec_parameters(ctx, &pubkey->u.ec.params);
+			if (r < 0)
+				free(pubkey->u.ec.ecpointQ.value);
 			LOG_TEST_RET(ctx, r, "Cannot fix EC parameters");
 		}
 	}
