@@ -165,6 +165,23 @@ static int openpgp_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 		key_info.u.ec.oid.value[key_info.u.ec.oid_len] = -1;
 		r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_STORE_KEY, &key_info);
 		break;
+	case SC_PKCS15_TYPE_PRKEY_EDDSA:
+		if (card->type != SC_CARD_TYPE_OPENPGP_GNUK) {
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "EdDSA keys not supported on this card");
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+		}
+		memset(&key_info, 0, sizeof(sc_cardctl_openpgp_keystore_info_t));
+		key_info.algorithm = (kinfo->id.value[0] == SC_OPENPGP_KEY_ENCR)
+				   ? SC_OPENPGP_KEYALGO_ECDH /* ECDH for slot 2 only */
+				   : SC_OPENPGP_KEYALGO_EDDSA; /* EdDSA for slot 1 and 3 */
+		key_info.key_id = kinfo->id.value[0];
+		/* TODO Test -- might not work */
+		key_info.u.ec.privateD = key->u.ec.privateD.data;
+		key_info.u.ec.privateD_len = key->u.ec.privateD.len;
+		key_info.u.ec.ecpointQ = key->u.ec.ecpointQ.value;
+		key_info.u.ec.ecpointQ_len = key->u.ec.ecpointQ.len;
+		r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_STORE_KEY, &key_info);
+		break;
 	default:
 		r = SC_ERROR_NOT_SUPPORTED;
 		sc_log(card->ctx, "%s: Key generation failed: Unknown/unsupported key type.", strerror(r));
@@ -367,6 +384,13 @@ static int openpgp_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card
 		}
 		r = openpgp_generate_key_ec(card, obj, pubkey);
 		break;
+	case SC_PKCS15_TYPE_PRKEY_EDDSA:
+		if (card->type != SC_CARD_TYPE_OPENPGP_GNUK) {
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "EdDSA is not supported on this card");
+			return SC_ERROR_NOT_SUPPORTED;
+		}
+		r = openpgp_generate_key_ec(card, obj, pubkey);
+		break;
 	default:
 		r = SC_ERROR_NOT_SUPPORTED;
 		sc_log(card->ctx, "%s: Key generation failed: Unknown/unsupported key type.", strerror(r));
@@ -419,6 +443,7 @@ static int openpgp_store_data(struct sc_pkcs15_card *p15card, struct sc_profile
 	case SC_PKCS15_TYPE_CERT:
 		cinfo = (sc_pkcs15_cert_info_t *) obj->data;
 		cid = &(cinfo->id);
+		unsigned int tag = 0x7F21;
 
 		if (cid->len != 1) {
 			sc_log(card->ctx, "ID=%s is not valid.", sc_dump_hex(cid->value, cid->len));
@@ -426,22 +451,37 @@ static int openpgp_store_data(struct sc_pkcs15_card *p15card, struct sc_profile
 		}
 
 		/* OpenPGP card v.2 contains only 1 certificate */
-		if (cid->value[0] != 3) {
+		if (cid->value[0] != 3 && p15card->card->type < SC_CARD_TYPE_OPENPGP_V3) {
 			sc_log(card->ctx,
 			       "This version does not support certificate ID = %d (only ID=3 is supported).",
 			       cid->value[0]);
 			LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
 		}
-		/* Just update the certificate DO */
+
+		/* OpenPGP card < v.3 does not support SELECT DATA calls */
+		if (p15card->card->type >= SC_CARD_TYPE_OPENPGP_V3) {
+			/* Mapping [3..1] passed --id to [0..2] for param */
+			u8 param = (u8) (2 - (cid->value[0] - 1));
+			/* check for unsigned underflow */
+			if (param > 2) {
+				LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+			}
+
+			/* Just update the certificate DO */
+			r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_SELECT_DATA, &param);
+			LOG_TEST_RET(card->ctx, r, "Failed OpenPGP - select data");
+		}
+
 		sc_format_path("7F21", path);
 		r = sc_select_file(card, path, &file);
+
 		LOG_TEST_RET(card->ctx, r, "Cannot select cert file");
 		r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_UPDATE);
 		sc_log(card->ctx,
 		       "Data to write is %"SC_FORMAT_LEN_SIZE_T"u long",
 		       content->len);
 		if (r >= 0 && content->len)
-			r = sc_put_data(p15card->card, 0x7F21, (const unsigned char *) content->value, content->len);
+			r = sc_put_data(p15card->card, tag, (const unsigned char *) content->value, content->len);
 		break;
 
 	case SC_PKCS15_TYPE_DATA_OBJECT:
