@@ -33,20 +33,24 @@
 
 #define MAX_MECHS 200
 
-#ifndef NDEBUG
-	#define debug_print(fmt, ...) \
-		{ fprintf(stderr, fmt "\n", ##__VA_ARGS__); } while (0)
-#else
-	#define debug_print(fmt, ...)
-#endif
+#define debug_print(fmt, ...) \
+	do { \
+		if (debug_flag) { \
+			fprintf(stderr, fmt "\n", ##__VA_ARGS__); \
+		} \
+	} while (0)
 
-#define FLAGS_SIGN		0x01
-#define FLAGS_SIGN_OPENSSL	0x02
+#define FLAGS_SIGN		0x001
+#define FLAGS_SIGN_OPENSSL	0x002
 #define FLAGS_SIGN_ANY		( FLAGS_SIGN | FLAGS_SIGN_OPENSSL )
-#define FLAGS_DECRYPT		0x04
-#define FLAGS_DECRYPT_OPENSSL	0x08
+#define FLAGS_DECRYPT		0x004
+#define FLAGS_DECRYPT_OPENSSL	0x008
 #define FLAGS_DECRYPT_ANY	( FLAGS_DECRYPT | FLAGS_DECRYPT_OPENSSL )
-#define FLAGS_DERIVE		0x10
+#define FLAGS_DERIVE		0x010
+#define FLAGS_WRAP		0x020
+#define FLAGS_WRAP_SYM		0x040
+#define FLAGS_UNWRAP		0x080
+#define FLAGS_UNWRAP_SYM	0x100
 
 typedef struct {
 	char *outfile;
@@ -59,9 +63,13 @@ typedef struct {
 
 typedef struct {
 	CK_MECHANISM_TYPE mech;
+	/* RSA-PSS parameters */
 	CK_MECHANISM_TYPE hash;
 	CK_RSA_PKCS_MGF_TYPE mgf;
 	int salt;
+	/* generic parameters used for example for secret keys */
+	void *params;
+	unsigned long params_len;
 	int usage_flags;
 	int result_flags;
 } test_mech_t;
@@ -76,15 +84,25 @@ typedef struct {
 	unsigned int interactive;
 	log_context_t log;
 
+	int verify_support;
+	int encrypt_support;
+
 	test_mech_t rsa_mechs[MAX_MECHS];
 	size_t  num_rsa_mechs;
 	test_mech_t	ec_mechs[MAX_MECHS];
 	size_t  num_ec_mechs;
+	test_mech_t	ed_mechs[MAX_MECHS];
+	size_t  num_ed_mechs;
+	test_mech_t	montgomery_mechs[MAX_MECHS];
+	size_t  num_montgomery_mechs;
+	test_mech_t	aes_mechs[MAX_MECHS];
+	size_t  num_aes_mechs;
 	test_mech_t	keygen_mechs[MAX_MECHS];
 	size_t  num_keygen_mechs;
 } token_info_t;
 
 extern token_info_t token;
+extern int debug_flag;
 
 #endif /* P11TEST_COMMON_H */
 
