@@ -21,13 +21,13 @@
 #include "config.h"
 #endif
 
-#ifdef ENABLE_OPENPACE
 #include "fread_to_eof.h"
 #include "npa-tool-cmdline.h"
 #include "sm/sm-eac.h"
 #include "sm/sslutil.h"
 #include "util.h"
 #include <eac/pace.h>
+#include <eac/objects.h>
 #include <libopensc/card-npa.h>
 #include <libopensc/log.h>
 #include <libopensc/opensc.h>
@@ -258,7 +258,11 @@ static int add_to_ASN1_AUXILIARY_DATA_NPA_TOOL(
 		goto err;
 	}
 
+#ifndef HAVE_EAC_OBJ_NID2OBJ
 	template->type = OBJ_nid2obj(nid);
+#else
+	template->type = EAC_OBJ_nid2obj(nid);
+#endif
 	if (!template->type) {
 		r = SC_ERROR_INTERNAL;
 		goto err;
@@ -858,10 +862,4 @@ err:
 
 	return -r;
 }
-#else
-int
-main (int argc, char **argv)
-{
-	return 1;
-}
-#endif
+
