@@ -46,12 +46,15 @@
 #include <openssl/err.h>
 #include <openssl/rand.h>
 #include <openssl/rsa.h>
-#include <openssl/dsa.h>
 #include <openssl/bn.h>
 #include <openssl/pkcs12.h>
 #include <openssl/x509v3.h>
 #include <openssl/crypto.h>
 #include <openssl/opensslconf.h> /* for OPENSSL_NO_EC */
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+# include <openssl/core_names.h>
+# include <openssl/param_build.h>
+#endif
 #ifndef OPENSSL_NO_EC
 #include <openssl/ec.h>
 #endif /* OPENSSL_NO_EC */
@@ -279,7 +282,7 @@ static const char *		option_help[] = {
 	"For a new key specify GUID for a MD container",
 	"Wait for card insertion",
 	"Display this message",
-	"Verbose operation. Use several times to enable debug output.",
+	"Verbose operation, may be used several times",
 
 	NULL,
 	NULL,
@@ -447,23 +450,8 @@ main(int argc, char **argv)
 {
 	struct sc_profile	*profile = NULL;
 	unsigned int		n;
-	int			r = 0;
-
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-	OPENSSL_config(NULL);
-#endif
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !(defined LIBRESSL_VERSION_NUMBER)
-	/* Openssl 1.1.0 magic */
-	OPENSSL_init_crypto(OPENSSL_INIT_LOAD_CRYPTO_STRINGS
-		| OPENSSL_INIT_ADD_ALL_CIPHERS
-		| OPENSSL_INIT_ADD_ALL_DIGESTS
-		| OPENSSL_INIT_LOAD_CONFIG,
-		NULL);
-#else
-	/* OpenSSL magic */
-	OpenSSL_add_all_algorithms();
-	OPENSSL_malloc_init();
-#endif
+	int					r = 0;
+	struct sc_pkcs15_card *tmp_p15_data = NULL;
 
 #ifdef RANDOM_POOL
 	if (!RAND_load_file(RANDOM_POOL, 32))
@@ -632,8 +620,13 @@ main(int argc, char **argv)
 	}
 
 out:
+	/* After erasing card profile->p15_data might change */
 	if (profile) {
+		tmp_p15_data = profile->p15_data;
 		sc_pkcs15init_unbind(profile);
+		if 	(tmp_p15_data != g_p15card) {
+ 			sc_pkcs15_unbind(tmp_p15_data);
+		}
 	}
 	if (g_p15card) {
 		sc_pkcs15_unbind(g_p15card);
@@ -727,7 +720,6 @@ static const struct alg_spec alg_types_sym[] = {
 
 static const struct alg_spec alg_types_asym[] = {
 	{ "rsa",	SC_ALGORITHM_RSA,	1024 },
-	{ "dsa",	SC_ALGORITHM_DSA,	1024 },
 	{ "gost2001",	SC_ALGORITHM_GOSTR3410,	SC_PKCS15_GOSTR3410_KEYSIZE },
 	{ "ec",		SC_ALGORITHM_EC,	0 },
 	{ NULL, -1, 0 }
@@ -1159,6 +1151,7 @@ is_cacert_already_present(struct sc_pkcs15init_certargs *args)
 
 	count = r;
 	for (i = 0; i < count; i++) {
+		int private_obj;
 		cinfo = (sc_pkcs15_cert_info_t *) objs[i]->data;
 
 		if (!cinfo->authority)
@@ -1168,7 +1161,8 @@ is_cacert_already_present(struct sc_pkcs15init_certargs *args)
 		/* XXX we should also match the usage field here */
 
 		/* Compare the DER representation of the certificates */
-		r = sc_pkcs15_read_certificate(g_p15card, cinfo, &cert);
+		private_obj = objs[i]->flags & SC_PKCS15_CO_FLAG_PRIVATE;
+		r = sc_pkcs15_read_certificate(g_p15card, cinfo, private_obj, &cert);
 		if (r < 0 || !cert)
 			continue;
 
@@ -1336,24 +1330,15 @@ do_read_check_certificate(sc_pkcs15_cert_t *sc_oldcert,
 	/* Compare the public keys, there's no high level openssl function for this(?) */
 	/* Yes there is in 1.0.2 and above EVP_PKEY_cmp */
 
-
 	r = SC_ERROR_INVALID_ARGUMENTS;
 	if (oldpk_type == newpk_type)
 	{
-#if  OPENSSL_VERSION_NUMBER >= 0x10002000L
-		if (EVP_PKEY_cmp(oldpk, newpk) == 1)
-			r = 0;
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L && !defined(LIBRESSL_VERSION_NUMBER)
+		if (EVP_PKEY_eq(oldpk, newpk) == 1)
 #else
-		if ((oldpk_type == EVP_PKEY_DSA) &&
-			!BN_cmp(EVP_PKEY_get0_DSA(oldpk)->p, EVP_PKEY_get0_DSA(newpk)->p) &&
-			!BN_cmp(EVP_PKEY_get0_DSA(oldpk)->q, EVP_PKEY_get0_DSA(newpk)->q) &&
-			!BN_cmp(EVP_PKEY_get0_DSA(oldpk)->g, EVP_PKEY_get0_DSA(newpk)->g))
-				r = 0;
-		else if ((oldpk_type == EVP_PKEY_RSA) &&
-			!BN_cmp(EVP_PKEY_get0_RSA(oldpk)->n, EVP_PKEY_get0_RSA(newpk)->n) &&
-			!BN_cmp(EVP_PKEY_get0_RSA(oldpk)->e, EVP_PKEY_get0_RSA(newpk)->e))
-				r = 0;
+		if (EVP_PKEY_cmp(oldpk, newpk) == 1)
 #endif
+		r = 0;
 	}
 
 	EVP_PKEY_free(newpk);
@@ -1383,6 +1368,7 @@ do_update_certificate(struct sc_profile *profile)
 	sc_pkcs15_cert_t *oldcert = NULL;
 	sc_pkcs15_der_t newcert_raw;
 	int r;
+	int private_obj;
 
 	if (opt_objectid == NULL) {
 		util_error("no ID given for the cert: use --id");
@@ -1401,7 +1387,8 @@ do_update_certificate(struct sc_profile *profile)
 		return r;
 
 	certinfo = (sc_pkcs15_cert_info_t *) obj->data;
-	r = sc_pkcs15_read_certificate(g_p15card, certinfo, &oldcert);
+	private_obj = obj->flags & SC_PKCS15_CO_FLAG_PRIVATE;
+	r = sc_pkcs15_read_certificate(g_p15card, certinfo, private_obj, &oldcert);
 	if (r < 0)
 		goto err;
 
@@ -1498,12 +1485,14 @@ static int get_cert_info(sc_pkcs15_card_t *myp15card, sc_pkcs15_object_t *certob
 	sc_pkcs15_object_t *otherobj;
 	sc_pkcs15_cert_t *othercert = NULL;
 	int r;
+	int private_obj;
 
 	*issuercert = NULL;
 	*has_sibling = 0;
 	*stop = 0;
 
-	r = sc_pkcs15_read_certificate(myp15card, (sc_pkcs15_cert_info_t *) certobj->data, &cert);
+	private_obj = certobj->flags & SC_PKCS15_CO_FLAG_PRIVATE;
+	r = sc_pkcs15_read_certificate(myp15card, (sc_pkcs15_cert_info_t *) certobj->data, private_obj, &cert);
 	if (r < 0)
 		return r;
 
@@ -1519,7 +1508,9 @@ static int get_cert_info(sc_pkcs15_card_t *myp15card, sc_pkcs15_object_t *certob
 			sc_pkcs15_free_certificate(othercert);
 			othercert=NULL;
 		}
-		r = sc_pkcs15_read_certificate(myp15card, (sc_pkcs15_cert_info_t *) otherobj->data, &othercert);
+
+		private_obj = otherobj->flags & SC_PKCS15_CO_FLAG_PRIVATE;
+		r = sc_pkcs15_read_certificate(myp15card, (sc_pkcs15_cert_info_t *) otherobj->data, private_obj, &othercert);
 		if (r < 0 || !othercert)
 			goto done;
 		if ((cert->issuer_len == othercert->subject_len) &&
@@ -1778,6 +1769,7 @@ do_generate_key(struct sc_profile *profile, const char *spec)
 	if (r == 0)
 		r = sc_pkcs15init_generate_key(g_p15card, profile, &keygen_args, keybits, NULL);
 	sc_unlock(g_p15card->card);
+	sc_pkcs15_free_prkey(&keygen_args.prkey_args.key);
 	return r;
 }
 
@@ -1793,13 +1785,14 @@ do_generate_skey(struct sc_profile *profile, const char *spec)
 
 	if ((r = init_skeyargs(&skey_args)) < 0)
 		return r;
-	skey_args.algorithm = algorithm;
 
 	algorithm = parse_alg_spec(alg_types_sym, spec, &keybits, 0);
 	if (algorithm < 0) {
 		util_error("Invalid symmetric key spec: \"%s\"", spec);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
+
+	skey_args.algorithm = algorithm;
 	skey_args.value_len = keybits;
 
 	r = sc_lock(g_p15card->card);
@@ -1869,16 +1862,27 @@ static void
 init_gost_params(struct sc_pkcs15init_keyarg_gost_params *params, EVP_PKEY *pkey)
 {
 #if !defined(OPENSSL_NO_EC)
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
 	EC_KEY *key;
+#else
+	char name[256];
+#endif
+	int nid = NID_undef;
 
 	assert(pkey);
 	if (EVP_PKEY_id(pkey) == NID_id_GostR3410_2001) {
+		assert(params);
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
 		key = EVP_PKEY_get0(pkey);
 		assert(key);
-		assert(params);
 		assert(EC_KEY_get0_group(key));
-		assert(EC_GROUP_get_curve_name(EC_KEY_get0_group(key)) > 0);
-		switch (EC_GROUP_get_curve_name(EC_KEY_get0_group(key))) {
+		nid = EC_GROUP_get_curve_name(EC_KEY_get0_group(key));
+#else
+		assert(EVP_PKEY_get_group_name(pkey, name, sizeof(name), NULL));
+		nid = OBJ_txt2nid(name);
+#endif
+		assert(nid > 0);
+		switch (nid) {
 		case NID_id_GostR3410_2001_CryptoPro_A_ParamSet:
 			params->gostr3410 = SC_PKCS15_PARAMSET_GOSTR3410_A;
 			break;
@@ -2953,14 +2957,8 @@ next: ;
 static void
 ossl_print_errors(void)
 {
-	static int	loaded = 0;
 	long		err;
 
-	if (!loaded) {
-		ERR_load_crypto_strings();
-		loaded = 1;
-	}
-
 	while ((err = ERR_get_error()) != 0)
 		fprintf(stderr, "%s\n", ERR_error_string(err, NULL));
 }
