@@ -65,11 +65,12 @@ int util_connect_reader (sc_context_t *ctx, sc_reader_t **reader,
 	}
 
 	if (do_wait) {
-		unsigned int event;
+		unsigned int event = 0;
 
 		if (sc_ctx_get_reader_count(ctx) == 0) {
 			fprintf(stderr, "Waiting for a reader to be attached...\n");
-			r = sc_wait_for_event(ctx, SC_EVENT_READER_ATTACHED, &found, &event, -1, NULL);
+			r = sc_wait_for_event(ctx, SC_EVENT_READER_ATTACHED|SC_EVENT_CARD_INSERTED,
+			                      &found, &event, -1, NULL);
 			if (r < 0) {
 				fprintf(stderr, "Error while waiting for a reader: %s\n", sc_strerror(r));
 				return r;
@@ -80,13 +81,17 @@ int util_connect_reader (sc_context_t *ctx, sc_reader_t **reader,
 				return r;
 			}
 		}
-		fprintf(stderr, "Waiting for a card to be inserted...\n");
-		r = sc_wait_for_event(ctx, SC_EVENT_CARD_INSERTED, &found, &event, -1, NULL);
-		if (r < 0) {
-			fprintf(stderr, "Error while waiting for a card: %s\n", sc_strerror(r));
-			return r;
+		if (event & SC_EVENT_CARD_INSERTED) {
+			*reader = found;
+		} else {
+			fprintf(stderr, "Waiting for a card to be inserted...\n");
+			r = sc_wait_for_event(ctx, SC_EVENT_CARD_INSERTED, &found, &event, -1, NULL);
+			if (r < 0) {
+				fprintf(stderr, "Error while waiting for a card: %s\n", sc_strerror(r));
+				return r;
+			}
+			*reader = found;
 		}
-		*reader = found;
 	}
 	else if (sc_ctx_get_reader_count(ctx) == 0) {
 		fprintf(stderr, "No smart card readers found.\n");
@@ -254,8 +259,8 @@ void util_hex_dump_asc(FILE *f, const u8 *in, size_t count, int addr)
 	}
 }
 
-NORETURN void
-util_print_usage_and_die(const char *app_name, const struct option options[],
+void
+util_print_usage(const char *app_name, const struct option options[],
 	const char *option_help[], const char *args)
 {
 	int i;
@@ -301,7 +306,13 @@ util_print_usage_and_die(const char *app_name, const struct option options[],
 		}
 		printf("  %-28s  %s\n", buf, option_help[i]);
 	}
+}
 
+NORETURN void
+util_print_usage_and_die(const char *app_name, const struct option options[],
+	const char *option_help[], const char *args)
+{
+	util_print_usage(app_name, options, option_help, args);
 	exit(2);
 }
 
