@@ -1,12 +1,13 @@
 #!/bin/bash
+SOURCE_PATH=${SOURCE_PATH:-..}
 
-source common.sh
+source $SOURCE_PATH/tests/common.sh
 
 echo "======================================================="
 echo "Setup SoftHSM"
 echo "======================================================="
 if [[ ! -f $P11LIB ]]; then
-    echo "WARNINIG: The SoftHSM is not installed. Can not run this test"
+    echo "WARNING: The SoftHSM is not installed. Can not run this test"
     exit 77;
 fi
 
