@@ -33,7 +33,7 @@
 #ifndef R_OK
 #define R_OK  4		/* test whether readable.  */
 #define W_OK  2		/* test whether writable.  */
-#define X_OK  1		/* test whether execubale. */
+#define X_OK  1		/* test whether executable. */
 #define F_OK  0		/* test whether exist.  */
 #endif
 
@@ -50,7 +50,7 @@
 #define ENABLE_PCSC
 #define HAVE_WINSCARD_H
 #ifndef DEFAULT_PCSC_PROVIDER
-#define DEFAULT_PCSC_PROVIDER "/System/Library/Frameworks/PCSC.framework/PCSC"
+#define DEFAULT_PCSC_PROVIDER "libpcsclite.so.1"
 #endif
 
 #define ENABLE_SHARED 1
@@ -61,7 +61,7 @@
 #define PATH_MAX FILENAME_MAX
 
 #ifndef PACKAGE_VERSION
-#define PACKAGE_VERSION "0.21.0"
+#define PACKAGE_VERSION "0.23.0"
 #endif
 
 #ifndef VERSION
@@ -81,7 +81,7 @@
 #endif
 
 #ifndef OPENSC_VERSION_MINOR
-#define OPENSC_VERSION_MINOR 21
+#define OPENSC_VERSION_MINOR 23
 #endif
 
 #ifndef OPENSC_VS_FF_COMPANY_NAME
@@ -113,7 +113,7 @@
 #endif
 
 #ifndef OPENSC_SCM_REVISION
-#define OPENSC_SCM_REVISION "OpenSC-0.21.0, rev: 3018098, commit-time: 2020-11-24 10:12:21 +0100"
+#define OPENSC_SCM_REVISION "OpenSC-<version not available>, rev: 5497519, commit-time: 2022-11-29 09:34:43 +0100"
 #endif
 
 #endif
